#include <math.h>
#include "barel3d.h"
#include "global.h"
#include "globmat.h"
#include "genfile.h"
#include "element.h"
#include "node.h"
#include "intpoints.h"
#include "loadcase.h"



/**
   JK, 27.9.2005
*/
barel3d::barel3d (void)
{
  long i;
  
  //  number nodes on element
  nne=2;
  //  number of DOFs on element
  ndofe=6;
  //  number of strain/stress components
  tncomp=1;
  //  number of functions approximated
  napfun=3;
  ssst = bar;
  
  //  number of blocks (parts of geometric matrix)
  nb=1;
  
  //  number of strain/stress components
  ncomp = new long [nb];
  ncomp[0]=1;
  
  //  cumulative number of components approximated
  cncomp = new long [nb];
  cncomp[0]=0;
  
  //  number of integration points
  //  order of numerical integration of stiffness matrix
  nip = new long* [nb];
  intordsm = new long* [nb];
  for (i=0;i<nb;i++){
    nip[i] = new long [nb];
    intordsm[i] = new long [nb];
  }
  
  nip[0][0]=1;

  //  total number of integration points
  tnip=0;
  for (i=0;i<nb;i++){
    tnip+=nip[i][i];
  }
  
  intordsm[0][0]=1;
}



barel3d::~barel3d (void)
{
  long i;
  
  for (i=0;i<nb;i++){
    delete [] nip[i];
    delete [] intordsm[i];
  }
  delete [] nip;
  delete [] intordsm;

  delete [] cncomp;
  delete [] ncomp;
}





/**
   Function computes direction %vector.
   
   @param s - direction %vector
   @param x,y,z - vectors of nodal coordinates
   
   @retval Function returns direction %vector s.

   JK, 28.9.2005
*/
void barel3d::dirvect (vector &s,double &l,vector &x,vector &y,vector &z)
{
  s[0]=x[1]-x[0];
  s[1]=y[1]-y[0];
  s[2]=z[1]-z[0];
  
  l=sqrt(s[0]*s[0]+s[1]*s[1]+s[2]*s[2]);
  
  if (l<Mp->zero){
    fprintf (stderr,"\n\n zero norm of direction vector in function dirvect (file %s, line %d).\n",__FILE__,__LINE__);
  }
  
  s[0]/=l;
  s[1]/=l;
  s[2]/=l;
}



/**
   function assembles strain-displacement (geometric) %matrix

   @param gm - strain-displacement (geometric) %matrix
   @param s - direction %vector
   @param l - length of the element

   @retval Function returns geometric %matrix in parameter gm. 

   JK, 27.9.2005
*/
void barel3d::geom_matrix (matrix &gm,vector &s,double l)
{
  gm[0][0]=s[0]/l*(-1.0);
  gm[0][1]=s[1]/l*(-1.0);
  gm[0][2]=s[2]/l*(-1.0);
  gm[0][3]=s[0]/l;
  gm[0][4]=s[1]/l;
  gm[0][5]=s[2]/l;
}



/**
   Function approximates function defined by nodal values.

   @param xi - coordinate on element
   @param nodval - nodal values
   
   @retval Function returns approximated value for the given natural coordinate xi.

   JK, 27.9.2005
*/
double barel3d::approx (double xi,vector &nodval)
{
  double f;
  vector bf(nne);

  bf_lin_1d (bf.a,xi);

  scprd (bf,nodval,f);

  return f;
}



/**
   function assembles transformation %matrix from local nodal coordinate
   system to the global coordinate system x_g = T x_l
   
   @param nodes - array containing node numbers
   @param tmat - transformation %matrix
   
   @retval Function returns transformation %matrix in parameter tmat.

   JK, 3.2.2002
*/
void barel3d::transf_matrix (ivector &nodes,matrix &tmat)
{
  long i;
  fillm (0.0,tmat);
  
  for (i=0;i<tmat.m;i++){
    tmat[i][i]=1.0;
  }
  
  for (i=0;i<nodes.n;i++){
    if (Mt->nodes[nodes[i]].transf>0){
      tmat[i*3+0][i*3]=Mt->nodes[nodes[i]].e1[0];
      tmat[i*3+1][i*3]=Mt->nodes[nodes[i]].e1[1];
      tmat[i*3+2][i*3]=Mt->nodes[nodes[i]].e1[2];
      
      tmat[i*3+0][i*3+1]=Mt->nodes[nodes[i]].e2[0];
      tmat[i*3+1][i*3+1]=Mt->nodes[nodes[i]].e2[1];
      tmat[i*3+2][i*3+1]=Mt->nodes[nodes[i]].e2[2];
      
      tmat[i*3+0][i*3+2]=Mt->nodes[nodes[i]].e3[0];
      tmat[i*3+1][i*3+2]=Mt->nodes[nodes[i]].e3[1];
      tmat[i*3+2][i*3+2]=Mt->nodes[nodes[i]].e3[2];
    }
  }  
}



/**
   Function assembles transformation %matrix T between coordinate system
   defined on element and global coordinate system.
      
   T x_local = x_global
   x_local = T^T x_global
   
   direction %vector denotes %vector defining local coordinate axis
   
   @param x - %vector of nodal coordinates in element coordinate system
   @param tmat - tranformation %matrix T
   @param gx,gy - vectors of nodal coordinates in global coordinate systems
   
*/
void barel3d::tran_mat (vector &x, matrix &tmat,vector &gx,vector &gy)
{
  long i,j;
  vector gv(2*gx.n),lv(2*gx.n);
  
  j=0;
  for (i=0;i<gx.n;i++){
    gv[j]=gx[i];  j++;
    gv[j]=gy[i];  j++;
  }
  
  //globloctransf (gv,lv,tmat);
  glvectortransf (gv,lv,tmat);

  for (i=0;i<gx.n;i++){
    x[i]=lv[i*2];
  }
}



/**
   Function computes stiffness %matrix of threedimensonal bar element.
   
   @param eid - element id
   @param ri,ci - row and column indices
   @param sm - stiffness %matrix

   JK, 27.9.2005
*/
void barel3d::stiffness_matrix (long eid,long ri,long ci,vector &s,matrix &sm)
{
  long ipp;
  double l,a,jac;
  vector x(nne),y(nne),z(nne);
  matrix gm(tncomp,ndofe),d(tncomp,tncomp);
  
  fillm (0.0,sm);
  
  // node coordinates
  Mt->give_node_coord3d (x,y,z,eid);
  
  //  length of the element
  l=sqrt ((x[1]-x[0])*(x[1]-x[0]) + (y[1]-y[0])*(y[1]-y[0]) + (z[1]-z[0])*(z[1]-z[0]));
  if (l<Mp->zero)
    print_err("zero length of the truss element",__FILE__,__LINE__,__func__);
  
  
  //  strain-displacement matrix
  geom_matrix (gm,s,l);
  //  integration point id
  ipp=Mt->elements[eid].ipp[ri][ci];
  //  stiffness matrix of material
  Mm->matstiff (d,ipp);
  //  are of cross section
  Mc->give_area (eid,a);
  jac=a*l;

  bdbj (sm.a,gm.a,d.a,jac,gm.m,gm.n);
}



/**
   Function computes stiffness %matrix of one element. If it is required, nodal values are transformed to 
   the local coordinate systems.
   
   @param eid - element id
   @param sm - stiffness %matrix
   
   @retval Function returns resulting stiffness %matrix in parameter sm

   JK, 27.9.2005
*/
void barel3d::res_stiffness_matrix (long eid,matrix &sm)
{
  long transf;
  double l;
  ivector nodes(nne);
  vector x(nne),y(nne),z(nne),s(3);
  matrix tmat (ndofe,ndofe);
  
  //  node coordinates
  Mt->give_node_coord3d (x,y,z,eid);
  
  //  direction vector
  dirvect (s,l,x,y,z);
  
  
  stiffness_matrix (eid,0,0,s,sm);


  //  nodes on element
  Mt->give_elemnodes (eid,nodes);
  //  transformation of stiffness matrix from global coordinate system
  //  to the local nodal systems
  transf = Mt->locsystems (nodes);
  if (transf>0){
    transf_matrix (nodes,tmat);
    glmatrixtransf (sm,tmat);
  }
}



/**
   Function computes mass %matrix of threedimensional bar element.
   
   @param eid - element id
   @param mm - mass %matrix

   @retval Function returns resulting mass %matrix in parameter mm

   JK, 27.9.2005
*/
void barel3d::mass_matrix (long eid,matrix &mm)
{
  double l,a,rho;
  vector x(nne),y(nne),z(nne);
  
  fillm (0.0,mm);

  Mt->give_node_coord3d (x,y,z,eid);

  //  length of the element
  l=sqrt ((x[1]-x[0])*(x[1]-x[0])+(y[1]-y[0])*(y[1]-y[0]));
  if (l<Mp->zero){
    fprintf (stderr,"\n\n zero length of the %ld truss element",eid);
    fprintf (stderr,"\n in function barel3d::mass_matrix (%s, line %d).\n",__FILE__,__LINE__);
  }
  Mc->give_area (eid,a);
  Mc->give_densitye (eid,rho);
  
  mm[0][0]=a*rho*l*1.0/3.0;
  mm[0][1]=0.0;
  mm[0][2]=a*rho*l*1.0/6.0;
  mm[0][3]=0.0;

  mm[1][0]=0.0;
  mm[1][1]=a*rho*l*1.0/3.0;
  mm[1][2]=0.0;
  mm[1][3]=a*rho*l*1.0/6.0;

  mm[2][0]=a*rho*l*1.0/6.0;
  mm[2][1]=0.0;
  mm[2][2]=a*rho*l*1.0/3.0;
  mm[2][3]=0.0;

  mm[3][0]=0.0;
  mm[3][1]=a*rho*l*1.0/6.0;
  mm[3][2]=0.0;
  mm[3][3]=a*rho*l*1.0/3.0;
}



/**
   Function computes mass %matrix of one element. If it is required, nodal values are transformed to 
   the local coordinate systems.

   @param eid - number of element
   @param mm  - mass %matrix

   @retval Function returns resulting mass %matrix in parameter mm
   
   TKo 3.2009
*/
void barel3d::res_mass_matrix (long eid,matrix &mm)
{
  vector gx(nne),gy(nne),x(nne);
  long transf;

  Mt->give_node_coord2d (gx,gy,eid);  
  giveloccoord (gx,gy,x);

  mass_matrix (eid,mm);

  //  transformation of mass matrix
  //  (in the case of nodal coordinate systems)
  ivector nodes (nne);
  Mt->give_elemnodes (eid,nodes);
  transf = Mt->locsystems (nodes);
  if (transf>0)
  {
    matrix tmat (ndofe,ndofe);
    transf_matrix (nodes,tmat);
    glmatrixtransf (mm,tmat);
  }
}



/**
  Function returns coordinates of nodes at local element coordinate system.

  @param x  - %vector of nodal x-coordinates in global coordinate system
  @param y  - %vector of nodal y-coordinates in global coordinate system
  @param lx - %vector of nodal coordinates at local element coordinate system

  @retval Function returns %vector of nodal coordinates at parameter lx.
*/
void barel3d::giveloccoord(vector &x, vector &y,vector &lx)
{
  double l;

  lx[0] = 0.0;
  l = sqr(x[1] - x[0]) + sqr(y[1] - y[0]);
  l = sqrt(l);
  lx[1] = l;
}



/**
   Function computes strains at integration points.
   
   @param lcid - load case id
   @param eid - element id
   
   JK, 4.10.2005
*/
void barel3d::res_ip_strains (long lcid,long eid)
{
  long transf;
  double l;
  vector gx(nne),gy(nne),gz(nne),s(3),r(ndofe),aux(ndofe);
  ivector cn(ndofe),nodes(nne);
  matrix tmat(ndofe,ndofe);
  
  //  node coordinates
  Mt->give_node_coord3d (gx,gy,gz,eid);
  //  direction vector
  dirvect (s,l,gx,gy,gz);
  
  //  assembling of displacements defined on the element
  eldispl (lcid,eid,r.a);

  
  //  transformation of displacement vector
  //  (in the case of nodal coordinate systems)
  Mt->give_elemnodes (eid,nodes);
  transf = Mt->locsystems (nodes);
  if (transf>0){
    //  transformation matrix
    transf_matrix (nodes,tmat);
    lgvectortransf (aux,r,tmat);
    copyv (aux,r);
  }
  
  ip_strains (lcid,eid,0,0,s,r,l);
}



/**
   Function computes strains at integration points of element.

   @param lcid - load case id
   @param eid - element id
   @param ri - row index
   @param ci - column index
   @param s - direction %vector
   @param r - vector of nodal displacements
   @param l - length of the element
   
   JK, 3.10.2005
*/
void barel3d::ip_strains (long lcid,long eid,long ri,long ci,vector &s,vector &r,double l)
{
  long i,ipid;
  vector eps(tncomp);
  matrix gm(tncomp,ndofe);
  
  ipid=Mt->elements[eid].ipp[ri][ci];
  for (i=0;i<intordsm[0][0];i++){
    
    //  geometrix matrix
    geom_matrix (gm,s,l);
    //  strain computation
    mxv (gm,r,eps);
    
    //  strain storage
    Mm->storestrain (lcid,ipid,eps);
    ipid++;
  }
}



/**
   Function computes strains in nodes of element.

   @param lcid - load case id
   @param eid - element id
   @param ri,ci - row and column indices

   JK, 10.5.2002
*/
void barel3d::nod_strains_ip (long lcid,long eid,long ri,long ci)
{
  long i,j,ipid;
  ivector ipnum(nne),nod(nne);
  vector eps(tncomp);
  
  //  numbers of integration points closest to nodes
  //  (function is from the file GEFEL/ordering.cpp)
  ipid=Mt->elements[eid].ipp[ri][ci];
  nodip_bar (ipid,intordsm[0][0],ipnum);
  
  //  node numbers of the element
  Mt->give_elemnodes (eid,nod);
  
  for (i=0;i<nne;i++){
    //  strains at the closest integration point
    Mm->givestrain (lcid,ipnum[i],eps);
    
    //  storage of strains to the node
    j=nod[i];
    Mt->nodes[j].storestrain (lcid,0,eps);
  }
}



/**
   Function computes strains at strain points.
   
   @param lcid - load case id
   @param eid - element id
   @param ri,ci - row and column indices
   
   JK
*/
void barel3d::strains (long lcid,long eid,long ri,long ci)
{
  long i,naep,ncp,sid;
  vector coord,eps;

  switch (Mm->stra.tape[eid]){
  case nowhere:{
    break;
  }
  case intpts:{
    //allip_strains (lcid,eid,ri,ci);
    break;
  }
  case enodes:{
    nod_strains_ip (lcid,eid,ri,ci);
    break;
  }
  case userdefined:{
    //  number of auxiliary element points
    naep = Mm->stra.give_naep (eid);
    ncp = Mm->stra.give_ncomp (eid);
    sid = Mm->stra.give_sid (eid);
    allocv (ncp,eps);
    allocv (1,coord);
    for (i=0;i<naep;i++){
      Mm->stra.give_aepcoord (sid,i,coord);

      if (Mp->strainaver==0)
	//appval (coord[0],0,ncp,eps,stra);
      if (Mp->strainaver==1)
	//appstrain (lcid,eid,coord[0],0,ncp,eps);

      Mm->stra.storevalues(lcid,eid,i,eps);
    }
    destrv (eps);
    destrv (coord);
    break;
  }
  default:
    print_err("unknown strain point is required",__FILE__,__LINE__,__func__);
  }

}



/**
   Function computes stresses at integration points.
   
   @param lcid - load case id
   @param eid - element id
   
   JK
*/
void barel3d::res_ip_stresses (long lcid,long eid)
{
  ip_stresses (lcid,eid,0,0);
}



/**
   Function computes stresses at integration points of element.

   @param eid - element id
   @param ri - row index
   @param ci - column index

   JK, 10.5.2002
*/
void barel3d::ip_stresses (long lcid,long eid,long ri,long ci)
{
  long ipid;
  
  ipid=Mt->elements[eid].ipp[ri][ci];
  
  //  computation of correct stresses
  if (Mp->strcomp==1)
    Mm->computenlstresses (ipid);
}



/**
   Function computes stresses at integration points of element.

   @param eid - element id
   @param ri - row index
   @param ci - column index

   JK, 10.5.2002
*/
void barel3d::ip_elast_stresses (long lcid,long eid,long ri,long ci)
{
  long i,ipid;
  vector eps(tncomp),sig(tncomp);
  matrix d(tncomp,tncomp);

  ipid=Mt->elements[eid].ipp[ri][ci];
  
  for (i=0;i<intordsm[0][0];i++){
    //  stiffness matrix of material
    Mm->matstiff (d,ipid);
    //  strain
    Mm->givestrain (lcid,ipid,eps);
    //  stress computation
    mxv (d,eps,sig);
    //  storage of stress
    Mm->storestress (lcid,ipid,sig);
    ipid++;
  }
}



/**
   Function computes stresses at nodes of element.

   @param lcid - load case id
   @param eid - element id
   @param ri,ci - row and column indices
   
   10.5.2002
*/
void barel3d::nod_stresses_ip (long lcid,long eid,long ri,long ci)
{
  long i,j,ipid;
  ivector ipnum(nne),nod(nne);
  vector sig(tncomp);
  
  //  numbers of integration points closest to nodes
  //  (function is from the file GEFEL/ordering.cpp)
  ipid=Mt->elements[eid].ipp[ri][ci];
  nodip_bar (ipid,intordsm[0][0],ipnum);
  
  //  node numbers of the element
  Mt->give_elemnodes (eid,nod);
  
  for (i=0;i<nne;i++){
    //  stresses at the closest integration point
    Mm->givestress (lcid,ipnum[i],sig);
    
    //  storage of stresses to the node
    j=nod[i];
    Mt->nodes[j].storestress (lcid,0,sig);
  }
}



/**
   Function computes stresses at required positions.
   
   @param lcid - load case id
   @param eid - element id
   @param ri,ci - row and column indices

*/
void barel3d::stresses (long lcid,long eid,long ri,long ci)
{
  long i,naep,ncp,sid;
  vector coord,sig;

  switch (Mm->stre.tape[eid]){
  case nowhere:{
    break;
  }
  case intpts:{
    //allip_stresses (lcid,eid,ri,ci);
    break;
  }
  case enodes:{
    nod_stresses_ip (lcid,eid,ri,ci);
    break;
  }
  case userdefined:{
    //  number of auxiliary element points
    naep = Mm->stre.give_naep (eid);
    ncp = Mm->stre.give_ncomp (eid);
    sid = Mm->stre.give_sid (eid);
    allocv (ncp,sig);
    allocv (1,coord);
    for (i=0;i<naep;i++){
      Mm->stre.give_aepcoord (sid,i,coord);

      if (Mp->stressaver==0)
	//appval (coord[0],0,ncp,sig,stre);
      if (Mp->stressaver==1)
	//appstress (lcid,eid,coord[0],0,ncp,sig);

      Mm->stre.storevalues(lcid,eid,i,sig);
    }
    destrv (sig);
    destrv (coord);
    break;
  }
  default:
    print_err("unknown stress point is required",__FILE__,__LINE__,__func__);
  }
}



/**
   Function computes other values in nodes of element.

   @param lcid - load case id
   @param eid - element id
   
   10.5.2002
*/
void barel3d::nod_eqother_ip (long lcid,long eid,long ri,long ci)
{
  long i,j,ipid,ncompo;
  ivector ipnum(nne),nod(nne);
  vector eqother;
  
  //  numbers of integration points closest to nodes
  //  (function is from the file GEFEL/ordering.cpp)
  ipid=Mt->elements[eid].ipp[ri][ci];
  nodip_bar (ipid,intordsm[0][0],ipnum);
  
  //  node numbers of the element
  Mt->give_elemnodes (eid,nod);
  
  for (i=0;i<nne;i++){
    ncompo = Mm->givencompeqother (ipnum[i],0);
    allocv (ncompo,eqother);
    Mm->giveeqother (ipnum[i],0,ncompo,eqother.a);
    
    //  storage of other values to the node
    j=nod[i];
    Mt->nodes[j].storeother (lcid,0,ncompo,eqother);
    
    destrv (eqother);
  }
}



/**
   Function assembles coordinates of integration points
   
   @param eid - element id
   @param ipp - integration point pointer
   @param ri - row index
   @param ci - column index
   @param coord - array containing coordinates of integration points
   
   @retval Function returns coordinates of integration points in the %vector coord.

   TKo, 7.2008
*/
void barel3d::ipcoord (long eid,long ipp,long ri,long ci,vector &coord)
{
  long ii,jj,i,aipp;
  double xi;
  vector w,gp;
  vector x(nne),y(nne),z(nne);

  Mt->give_node_coord3d (x,y,z,eid);
  for (ii = 0; ii < nb; ii++){
    for (jj = 0; jj < nb; jj++){
      //  integration point id
      aipp=Mt->elements[eid].ipp[ri+ii][ci+jj];
      if (intordsm[ii][jj] == 0)
        continue;
      allocv (intordsm[ii][jj],gp);
      allocv (intordsm[ii][jj],w);
      gauss_points (gp.a,w.a,intordsm[ii][jj]);
      
      for (i=0;i<intordsm[ii][jj];i++){
        xi=gp[i];
        if (aipp==ipp){
          coord[0]=approx (xi,x);
          coord[1]=approx (xi,y);
          coord[2]=approx (xi,z);
        }
        aipp++;
      }
      destrv (gp);  destrv (w);
    }
  }
}



/**
   Function computes volume appropriate to integration point.

   @param eid - element id
   @param ri  - row index of the given integration point block
   @param ci  - column index of the given integration point block

   TKo, 7.2008
*/
void barel3d::ipvolume (long eid,long ri,long ci)
{
  long ipp;
  double l,a,vol;
  vector x(nne),y(nne),z(nne);
  
  // node coordinates
  Mt->give_node_coord3d (x,y,z,eid);
  
  //  length of the element
  l=sqrt ((x[1]-x[0])*(x[1]-x[0]) + (y[1]-y[0])*(y[1]-y[0]) + (z[1]-z[0])*(z[1]-z[0]));
  if (l<Mp->zero){
    char emsg[100];
    sprintf(emsg, "zero length of the %ld truss element", eid);
    print_err(emsg, __FILE__, __LINE__, __func__);
  }
  //  are of cross section
  Mc->give_area (eid,a);
  // computation of volume
  vol=a*l;
  //  integration point id
  ipp=Mt->elements[eid].ipp[ri][ci];
  Mm->storeipvol (ipp,vol);
}



/**
   Function interpolates the nodal values to the integration points on the element
   
   @param eid    - element id
   @param nodval - nodal values
   @param ipval  - value at integration points
   
   @retval Function returns approximated values at integration points in the %vector ipval.
   
   JK, 3.10.2005
*/
void barel3d::intpointval (long eid,vector &nodval,vector &ipval)
{
  long ii,jj,i,k,ipp;
  double xi;
  vector w,gp;
  
  k=0;
  for (ii = 0; ii < nb; ii++){
    for (jj = 0; jj < nb; jj++){
      
      //  integration point id
      ipp=Mt->elements[eid].ipp[ii][jj];
      if (intordsm[ii][jj] == 0)
        continue;
      allocv (intordsm[ii][jj],gp);
      allocv (intordsm[ii][jj],w);
      gauss_points (gp.a,w.a,intordsm[ii][jj]);
      
      for (i=0;i<intordsm[ii][jj];i++){
        xi=gp[i];
        //  value at integration point
        ipval[k] = approx (xi,nodval);
        k++;
        ipp++;
      }
      destrv (gp);  destrv (w);
    }
  }
}



/**
  The function computes initial values of the given quantities at each integration point of the
  element from the nodal values given by the parameter nodval. Initial condition types must be 
  the same for all nodes of the element.

  @param eid - element id
  @param ri  - block row index
  @param ci  - block column index
  @param nodval - nodal values of particular initial conditions.
                  nodval[i][j] represents value of j-th initial condition at i-th node of the given element.
  @param ictn - array of types of initial condition for each node of element.
                The type of initial condition determines which values are being specified in the node. 
                (ictn[i] & inistrain) returns nonzero if nodal values of initial strains are specified
                (ictn[i] & inistress) returns nonzero if nodal values of initial stresses are specified
                (ictn[i] & iniother)  returns nonzero if nodal values of initial values of eqother array are specified
                (ictn[i] & inicond)   returns nonzero if nodal values of other initial conditions are specified

  @retur The function does not return anything.

  Created by Tomas Koudelka 2004
  Revised by Tomas Koudelka 03.2012
*/
void barel3d::inicipval(long eid, long ri, long ci, matrix &nodval, inictype *ictn)
{
  long i, j, k, ipp;
  long ii, jj, nv = nodval.n;
  double xi, ipval;
  vector w, gp, anv(nne);
  long nstra, nstre, ncompstr, ncompeqother;
  long idstra, idstre, idoth, idic;
  inictype ict;
  int aux;

  nstra = idstra = nstre = idstre = idoth = idic = 0;

  ict = ictn[0];
  for (i=0; i<nne; i++)
  {
    aux = int(ictn[i])-int(ict);
    if (aux < 0)  aux = -aux;    
    aux &= ~(inidisp);
    aux &= ~(inidisp_x);
    aux &= ~(inidisp_y);
    aux &= ~(inidisp_z);
    if ((ictn[i] != ict) && aux)
    {
      print_err("Incompatible types of initial conditions on element %ld\n"
                " at %ld. and %ld. nodes", __FILE__, __LINE__, __func__, eid+1, 1, i+1);
      abort();
    }
  }
  for (j = 0; j < nv; j++) // for all initial values
  {
    for(i = 0; i < nne; i++) // for all nodes on element
      anv[i] = nodval[i][j];
    for (ii = 0; ii < nb; ii++)
    {
      for (jj = 0; jj < nb; jj++)
      {
        ipp=Mt->elements[eid].ipp[ri+ii][ci+jj];
        if (intordsm[ii][jj] == 0)
          continue;
        allocv (intordsm[ii][jj],gp);
        allocv (intordsm[ii][jj],w);
        gauss_points (gp.a,w.a,intordsm[ii][jj]);
        for (k = 0; k < intordsm[ii][jj]; k++)
        {
          xi=gp[k];
          //  value in integration point
          ipval = approx (xi,anv);
          if ((ictn[i] & inistrain) && (j < Mm->ip[ipp].ncompstr))
          {
            Mm->ip[ipp].strain[idstra] += ipval;
            ipp++;
            continue;
          }
          if ((ictn[i] & inistress) && (j < nstra + Mm->ip[ipp].ncompstr))
          {
            Mm->ip[ipp].stress[idstre] += ipval;
            ipp++;
            continue;
          }
          if ((ictn[0] & iniother) && (j < nstra+nstre+ncompeqother))
          {
            Mm->ip[ipp].eqother[idoth] += ipval;
            ipp++;
            continue;
          }
          if ((ictn[0] & inicond) && (j < nv))
          {
            if (Mm->ic[ipp] == NULL)
            {
              Mm->ic[ipp] = new double[nv-j];
              memset(Mm->ic[ipp], 0, sizeof(*Mm->ic[ipp])*(nv-j)); 
	    }
            Mm->ic[ipp][idic] += ipval;
            ipp++;
            continue;
          }
          ipp++;
        }
        destrv (gp);  destrv (w);
      }
    }
    ipp=Mt->elements[eid].ipp[ri][ci];
    ncompstr =  Mm->ip[ipp].ncompstr;
    ncompeqother = Mm->ip[ipp].ncompeqother;
    if ((ictn[0] & inistrain) && (j < ncompstr))
    {
      nstra++;
      idstra++;
      continue;
    }
    if ((ictn[0] & inistress) && (j < nstra + ncompstr))
    {      
      nstre++;
      idstre++;
      continue;
    }  
    if ((ictn[0] & iniother)  && (j < nstra + nstre + ncompeqother))
    {
      idoth++;
      continue;
    }
    if ((ictn[0] & inicond) && (j < nv))
    {
      idic++;
      continue;
    }
  }
}



/**
   Function computes internal forces.
   
   @param lcid - load case id
   @param eid - element id
   @param ri,ci - row and column indices
   @param s - direction %vector
   @param ifor - %vector of internal forces
   
   @retval Function returns nodal values of internal forces in %vector ifor.

   JK, 3.10.2005
*/
void barel3d::internal_forces (long lcid,long eid,long ri,long ci,vector &ifor)
{
  integratedquant iq;
  iq=locstress;
  vector x(nne),y(nne),z(nne);

  //  computation of stresses
  compute_nlstress (lcid,eid,ri,ci);
    
  Mt->give_node_coord3d (x,y,z,eid);

  //  integration of stresses over the element
  elem_integration (iq,lcid,eid,ri,ci,ifor,x,y,z);
}



/**
   Function computes internal forces for nonlocal models


   @param lcid - number of load case
   @param eid - element id
   @param ri,ci - row and column indices
   @param ifor - %vector of internal forces
   
   @retval Function returns nodal values of nonlocal internal forces in %vector ifor.

   TKo, 7.2008
*/
void barel3d::nonloc_internal_forces (long lcid,long eid,long ri,long ci,vector &ifor)
{
  integratedquant iq=nonlocstress;
  vector x(nne),y(nne),z(nne);
  Mt->give_node_coord3d (x,y,z,eid);
  
  //  computation of nonlocal stresses
  compute_nonloc_nlstress (lcid,eid,ri,ci);
  
  //  integration of stresses over the element
  elem_integration (iq,lcid,eid,ri,ci,ifor,x,y,z);
}



/**
   Function computes increment of internal forces.
   
   @param lcid - load case id
   @param eid - element id
   @param ri,ci - row and column indices
   @param ifor - vector of internal forces
   
   @retval Function returns nodal values of increments of internal forces in %vector ifor.

   JK, 29.4.2008
*/
void barel3d::incr_internal_forces (long lcid,long eid,long ri,long ci,vector &ifor)
{
  integratedquant iq;
  iq=stressincr;
  vector x(nne),y(nne),z(nne);

  //  computation of stresses
  compute_nlstressincr (lcid,eid,ri,ci);
  
  Mt->give_node_coord3d (x,y,z,eid);

  //  integration of stresses over the element
  elem_integration (iq,lcid,eid,ri,ci,ifor,x,y,z);

}



/**
   Function computes nodal forces caused by temperature changes.
   
   @param eid - element id
   @param ri,ci - row and column indices
   @param nfor - array containing nodal forces
   
   @retval Function returns nodal values of forces caused by eigenstrains in %vector nfor.

   30.11.2002, JK
       7.2008, TKo
*/
void barel3d::eigstrain_forces (long lcid,long eid,long ri,long ci,vector &nfor)
{
  integratedquant iq;
  iq=eigstress;
  vector x(nne),y(nne),z(nne);
  
  Mt->give_node_coord3d (x,y,z,eid);
  
  //  computation of eigenstresses
  if (Mp->eigstrcomp)
    compute_eigstress (lcid,eid,ri,ci);
  
  //  integration of stresses over the element
  elem_integration (iq,lcid,eid,ri,ci,nfor,x,y,z);
}



/**
   Function computes resulting internal forces. If it is required, nodal values are transformed to 
   the local coordinate systems.

   @param lcid - load case id
   @param eid  - element id
   @param ifor - %vector of internal forces at nodes

   @retval Function returns resulting internal forces in parameter ifor.
   
   JK, 3.10.2005
*/
void barel3d::res_internal_forces (long lcid,long eid,vector &ifor)
{
  long transf;
  ivector nodes(nne);

  internal_forces (lcid,eid,0,0,ifor);

  //  nodes on element
  Mt->give_elemnodes (eid,nodes);
  //  transformation of stiffness matrix from global coordinate system
  //  to the local nodal systems
  transf = Mt->locsystems (nodes);
  if (transf>0){
    matrix tmat (ndofe,ndofe);
    vector v(ndofe);
    transf_matrix (nodes,tmat);
    glvectortransf (ifor,v,tmat);
    copyv (v,ifor);
  }
}



/**
   Function computes resulting increment of nonlocal internal forces.

   @param lcid - load case id
   @param eid  - element id
   @param ifor - %vector of internal forces at nodes

   @retval Function returns resulting nonlocal internal forces in parameter ifor.
   
   TKo 7.2008
*/
void barel3d::res_nonloc_internal_forces (long lcid,long eid,vector &ifor)
{
  long transf;
  ivector nodes(nne);

  nonloc_internal_forces (lcid,eid,0,0,ifor);

  //  nodes on element
  Mt->give_elemnodes (eid,nodes);
  //  transformation of stiffness matrix from global coordinate system
  //  to the local nodal systems
  transf = Mt->locsystems (nodes);
  if (transf>0){
    matrix tmat (ndofe,ndofe);
    vector v(ndofe);
    transf_matrix (nodes,tmat);
    glvectortransf (ifor,v,tmat);
    copyv (v,ifor);
  }
}



/**
   Function computes resulting increments of internal forces.

   @param lcid - load case id
   @param eid  - element id
   @param ifor - %vector of internal forces at nodes

   @retval Function returns resulting increments of internal forces in parameter ifor.
   
   29.4.2008
*/
void barel3d::res_incr_internal_forces (long lcid,long eid,vector &ifor)
{
  long transf;
  ivector nodes(nne);

  incr_internal_forces (lcid,eid,0,0,ifor);

  //  nodes on element
  Mt->give_elemnodes (eid,nodes);
  //  transformation of stiffness matrix from global coordinate system
  //  to the local nodal systems
  transf = Mt->locsystems (nodes);
  if (transf>0){
    matrix tmat (ndofe,ndofe);
    vector v(ndofe);
    transf_matrix (nodes,tmat);
    glvectortransf (ifor,v,tmat);
    copyv (v,ifor);
  }
}



/**
   Function computes resulting internal forces caused by eigenstrains.

   @param lcid - load case id
   @param eid  - element id
   @param nfor - %vector of nodal forces caused by eigenstrains

   @retval Function returns resulting increments of internal forces in parameter nfor.

   12.8.2001
*/
void barel3d::res_eigstrain_forces (long lcid,long eid,vector &nfor)
{
  long transf;
  ivector nodes(nne);

  eigstrain_forces (lcid,eid,0,0,nfor);

  //  nodes on element
  Mt->give_elemnodes (eid,nodes);
  //  transformation of stiffness matrix from global coordinate system
  //  to the local nodal systems
  transf = Mt->locsystems (nodes);
  if (transf>0){
    matrix tmat (ndofe,ndofe);
    vector v(ndofe);
    transf_matrix (nodes,tmat);
    glvectortransf (nfor,v,tmat);
    copyv (v,nfor);
  }
}



/**
   Function computes correct stresses at integration points on element.

   @param lcid - number of load case
   @param eid - element id
   @param ri,ci - row and column indices
   
   JK, 1.12.2006
*/
void barel3d::compute_nlstress (long lcid,long eid,long ri,long ci)
{
  long ipid;
  
  ipid=Mt->elements[eid].ipp[ri][ci];
  
  //  computation of correct stresses
  if (Mp->strcomp==1)
    Mm->computenlstresses (ipid);
  
}



/**
   Function computes local values which will be used for averageing 
   in nonlocal models at integration points. Mp->nonlocphase have to be 1.

   @param lcid - number of load case
   @param eid - element id
   @param ri,ci - row and column indices
   
   TKo, 7.2008
*/
void barel3d::local_values (long lcid,long eid,long ri,long ci)
{
  long ipp;
  
  ipp=Mt->elements[eid].ipp[ri][ci];
  
  
  //  computation of correct stresses
  if (Mp->strcomp==1)
    Mm->computenlstresses (ipp);
  
}



/**
   Function computes nonlocal correct stresses at integration points on element.
   
   @param lcid - number of load case
   @param eid - element id
   @param ri,ci - row and column indices
   
   JK, 1.12.2006
*/
void barel3d::compute_nonloc_nlstress (long lcid,long eid,long ri,long ci)
{
  long ipid;
  
  ipid=Mt->elements[eid].ipp[ri][ci];

  //  computation of correct stresses
  if (Mp->strcomp==1)
    Mm->compnonloc_nlstresses (ipid);
  
}



/**
   Function computes correct stress increments at integration points on element.

   @param lcid - number of load case
   @param eid - element id
   @param ri,ci - row and column indices
   
   TKo, 7.2008
*/
void barel3d::compute_nlstressincr (long lcid,long eid,long ri,long ci)
{
  long ipid;
  
  ipid=Mt->elements[eid].ipp[ri][ci];
  
  //  computation of correct stresses
  if (Mp->strcomp==1)
    Mm->computenlstressesincr (ipid);
  
}



/**
   Function computes correct stresses caused by eigenstrains at integration points on element.
   
   @param lcid - number of load case
   @param eid - element id
   @param ri,ci - row and column indices
   
   JK, 1.12.2006
*/
void barel3d::compute_eigstress (long lcid,long eid,long ri,long ci)
{
  long ipid,rmid;
  vector eigstr(ASTCKVEC(tncomp)),sig(ASTCKVEC(tncomp));
  matrix d(ASTCKMAT(tncomp,tncomp));
  
  //  integration point number
  ipid=Mt->elements[eid].ipp[ri][ci];
  //  index of relaxation model in tm array
  rmid = Mm->ip[ipid].grmid ();
  
  if (rmid>-1){
    //  eigenstresses are computed from a given formula, not from eigstrains
    Mm->eigenstresses (sig,ipid,rmid);
  }else{
    //
    //  eigenstresses are computed from eigenstrains \sigma_0 = D (-\eps_0)
    //
    // restore eigenstrains
    Mm->giveeigstrain (ipid,eigstr);
    // change sign of eigenstrain vector
    chsgnv(eigstr);
    //  matrix of stiffness of the material
    Mm->matstiff (d,ipid);
    // calculate eigenstresses    
    mxv (d,eigstr,sig);
  }
  
  Mm->storeeigstress (ipid,sig);
  
}

/**
   Function integrates selected quantity over the finite element.
   
   @param iq - type of integrated quantity (see alias.h)
   @param lcid - number of load case
   @param eid - element id
   @param ri,ci - row and column indices
   @param nv - nodal values
   @param x,y,z - node coordinates
   
   @retval Function returns integrated values at nodes in %vector nv.

   JK, 1.12.2006
*/
void barel3d::elem_integration (integratedquant iq,long lcid,long eid,long ri,long ci,vector &nv,vector &x,vector &y,vector &z)
{
  long ipp;
  double l,a;
  ivector nodes(nne);
  vector ipv(tncomp),contr(ndofe);
  vector s(3);
  matrix gm(tncomp,ndofe);
  
  //  direction vector
  dirvect (s,l,x,y,z);

  ipp=Mt->elements[eid].ipp[ri][ci];

  //  length of the element
  l=sqrt ((x[1]-x[0])*(x[1]-x[0]) + (y[1]-y[0])*(y[1]-y[0]) + (z[1]-z[0])*(z[1]-z[0]));
  if (l<Mp->zero)
    print_err("zero length of the truss element",__FILE__,__LINE__,__func__);

  //  function assembles required quantity at integration point
  Mm->givequantity (iq,lcid,ipp,cncomp[0],ipv);
  //  strain-displacement (geometric) matrix
  geom_matrix (gm,s,l);
  //  contribution to the nodal values
  mtxv (gm,ipv,contr);
  //  area of bar cross-section
  Mc->give_area (eid,a);
  cmulv (a*l,contr);
  //  summation
  addv(contr,nv,nv);
}

/**
   Function searches the minimum and maximum strain in integration points
   
   @param min - the minimum strain
   @param max - the maximum strain
   @param lcid - load case id
   @param eid - element id
   
   @retval Function returns integrated values at nodes in %vector nv.

   JK, 16. 11. 2012
*/
void barel3d::find_extreme_strains (vector &min,vector &max,long lcid,long eid)
{
  long i,ipp;
  vector eps(tncomp);
  
  //  integration point id
  ipp=Mt->elements[eid].ipp[0][0];
  
  //  strains
  Mm->givestrain (lcid,ipp,eps);
  
  //  loop over the number of strain components
  for (i=0;i<tncomp;i++){
    if (min[i]>eps[i])
      min[i]=eps[i];
    if (max[i]<eps[i])
      max[i]=eps[i];
  }//  end of the loop over the number of strain components
  
}

/**
   Function searches the minimum and maximum stresses in integration points
   
   @param min - the minimum stress
   @param max - the maximum stress
   @param lcid - load case id
   @param eid - element id
   
   JK, 16. 11. 2012
*/
void barel3d::find_extreme_stresses (vector &min,vector &max,long lcid,long eid)
{
  long i,ipp;
  vector sig(tncomp);
  
  //  integration point id
  ipp=Mt->elements[eid].ipp[0][0];
  
  //  strains
  Mm->givestress (lcid,ipp,sig);
  
  //  loop over the number of stress components
  for (i=0;i<tncomp;i++){
    if (min[i]>sig[i])
      min[i]=sig[i];
    if (max[i]<sig[i])
      max[i]=sig[i];
  }//  end of the loop over the number of stress components
  
}
