/*
    Copyright (C) 2003-2015  David Masin
    email: masin@natur.cuni.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation 
    59 Temple Place, Suite 330, Boston, MA, 02111-1307, USA

*/

/****************** Hypoplasticity for unsaturated soils, expansive soils with *******************/
/****************** double-porosity structure, thermal effects ***********************************/

#include "generalmod.h"
#include <string.h>
/**
  The function calculates increments of net stress, Sr and state variables dqstatev based on their 
  initial values and increments of strains, pore fluid pressures and temperature. Sign convention: compression negative for stresses, strains and pressures.
  
  int nparms nstatev, ngstrain, ngstress jsou proměnné struktury General_model inicializované automaticky při jejím vytvoření.

  @param strain_gen[ngstrain] - (input): controlled variables [eps11, eps22, eps33, gamma12, gamma13, gamma23, s, T]^T
  @param stress_gen[ngstress] - (input): implied variables, will not be updated [signet11, signet22, signet33, signet12, signet13, signet23, Sr]^T. "signet" is net stress, that is total stress minus krondelta x ua.
  @param qstatev[nstatev] - (input): state variable vector, will be updated only for flag==0.
  @param dstrain_gen[ngstrain] - (input): increment of controlled variables [deps11, deps22, deps33, dgamma12, dgamma13, dgamma23, ds, dT]^T
  @param dstress_gen[ngstress] - (output): increment of implied variables [dsig11, dsig22, dsig33, dsig12, dsig13, dsig23, dSr]^T
  @param dqstatev[nstatev] - (output): increment of state variables
  @param DDtan_gen[ngstress][ngstrain] - (output): generalised tangent stiffness d(dstress_gen)/d(dstrain_gen)
  @param parms[nparms] - (input): model parameters
  @param flag - (input): 0 for state variable initialisation, 1 for DDtan_gen return, 2 for standard run, 3 DDtan_gen, but only non-mechanical components of stiffness matrix are calculated.

  @return The function does not return anything.
 
  Created by D. Masin 13.7.2015
*/

long Hypoplasti_unsat_expansive_thermal::soil_model(double strain_gen[], double stress_gen[], double qstatev[], 
  double dstrain_gen[], double dstress_gen[], double dqstatev[], double *DDtan_gen, 
  double parms[], int flag) {

  //Definition of variables in full 3x3=9 tensorial notation, to be used for any flag
  long chyba = 0;
  double signet[9];
  double deps[9];
  double suction=0;
  double dsuction=0;
  double Temper=0;
  double dTemper=0;
  double Sr=0;
  double othervar[5]={0,0,0,0,0};
  convert_from_general(stress_gen, signet, othervar, (ngstress-6), 1);
  
  //Application of pt
  pt=parms[21];
  signet[0]-=pt;
  signet[4]-=pt;
  signet[8]-=pt;
  Sr=othervar[0];
    
  double ddum9[9];
  convert_from_general(strain_gen, ddum9, othervar, (ngstrain-6), 2);
  suction=othervar[0];
  Temper=othervar[1];
  
  bool debug=false;
  /*debug code
  long newchyba=test_math_err();
  if(newchyba) cout<<"newchyba"<<endl;
  errno=0;
  */

  if(debug) {
    cout<<flag<<endl;
    for(int i=0; i<ngstrain; i++) cout<<"strain_gen["<<i<<"]="<<strain_gen[i]<<endl;
    for(int i=0; i<ngstress; i++) cout<<"stress_gen["<<i<<"]="<<stress_gen[i]<<endl;        
    for(int i=0; i<nstatev; i++) cout<<"qstatev["<<i<<"]="<<qstatev[i]<<endl;
    if(dstrain_gen != NULL) for(int i=0; i<ngstrain; i++) cout<<"dstrain_gen["<<i<<"]="<<dstrain_gen[i]<<endl;
    if(dstress_gen != NULL) for(int i=0; i<ngstress; i++) cout<<"dstress_gen["<<i<<"]="<<dstress_gen[i]<<endl; 
    if(dqstatev != NULL) for(int i=0; i<nstatev; i++) cout<<"dqstatev["<<i<<"]="<<dqstatev[i]<<endl;
    for(int i=0; i<nparms; i++) cout<<"parms["<<i<<"]="<<parms[i]<<endl;
    
    cout<<"----------------------------end step-----------------------------------"<<endl;
  }
 
  //Initialise state variables and parameters
  if (flag == 0) {
      
    //start define parameters    
    phic=parms[0]*3.141592653589793238462/180;
    lambda=parms[1];
    kappa=parms[2];
    N=parms[3];
    nuparam=parms[4];
    Ocparam=2;

    nparam=parms[5];
    lparam=parms[6];
    nTparam=parms[7];
    lTparam=parms[8];
    mmparam=parms[9];//double structure coupling m
    mparam=10;//collapse m

    alpha_s=parms[10];
    kappam=parms[11];
    smstar=parms[12];
    emstar=parms[13];
    csh=parms[14];

    sairentry0=parms[15];
    eM0=parms[16];
    Trparam=parms[17];
    aTparam=parms[18];
    bTparam=parms[19];

    lambdap0=0.55;
    aer=parms[20];
    pt=parms[21];
    alphaG=1;
    alphanu=alphaG;
    alphaE=gscalar_power(alphaG,1./0.8);
  
    rlambdascan=1/10;
    //end define parameters    
     
    double zeroinctens[9]={0,0,0,0,0,0,0,0,0};
    double zeroincscal=0;
    
    double fu=0;
    long ret = update_hisv(signet, suction, Temper, qstatev, zeroinctens, zeroincscal, zeroincscal, zeroinctens, Sr, dqstatev, flag, fu);
    garray_add(qstatev, dqstatev, qstatev, nstatev);
    if(qstatev[5]<0) {
      cout<<"Initial value of eM is lower than 0 and this is not allowed, please re-initialise"<<endl;
      exit(0);
    }
    
    stress_gen[6]=Sr;
    return ret;
  }
  //Calculate DDtan_gen as L tensor for mechanical components, the other components found by perturbation from zero
  else if((flag==1) || (flag==3)){
    
    double pertinc_all[ngstrain];
    for(int i=0; i<6; i++) pertinc_all[i]=0;
    pertinc_all[6]=-0.0001; //ds
    pertinc_all[7]=0.0001; //dT
    garray_set(DDtan_gen, 0, ngstrain*ngstress);
    double multfactor=3;
    
    if(ngstrain>6) {
      int initi=6;
      if(ngstress<=6) initi=6;
      for(int i=initi; i<ngstrain; i++) {
    
        double strain_gen_pert[ngstrain];
        garray_set(strain_gen_pert, 0, ngstrain);
        strain_gen_pert[i]=pertinc_all[i];
                
        double deps_pert[9];
        double dsignet_pert[9];
        garray_set(dsignet_pert,0,9);
        garray_set(deps_pert,0,9);
        
        convert_from_general(strain_gen_pert, deps_pert, othervar, (ngstrain-6), 2);
        double dsuction_pert=othervar[0];
        double dTemper_pert=othervar[1];
        double Sr_pert=0;
        double dqstatev_pert[nstatev];

        long thischyba=calc_fsigq(signet, suction, Temper, qstatev, deps_pert, dsuction_pert, dTemper_pert, dsignet_pert, Sr_pert, dqstatev_pert);

        if(!chyba) chyba=thischyba;
      
        double dsig_gen_pert[ngstress];
        othervar[0]=Sr_pert-Sr;
        convert_to_general(dsig_gen_pert, dsignet_pert, othervar, (ngstress-6), 1);
           
        for(int j=0; j<ngstress; j++) {
          DDtan_gen[j*ngstrain+i]=multfactor*dsig_gen_pert[j]/strain_gen_pert[i];
        }
      }
    }
    if (flag==3)
      return(0);    
    
    double hypo_L[3][3][3][3];
    double DDtan[3][3][3][3];
    double DDtan_abq[6][6];
    double hypo_N[3][3];
    double H_unsat[3][3];
    double HT_unsat[3][3];
    garray_set(&hypo_L[0][0][0][0],0,81);
    garray_set(&DDtan[0][0][0][0],0,81);
    garray_set(&DDtan_abq[0][0],0,36);
    garray_set(&hypo_N[0][0],0,9);
    garray_set(&H_unsat[0][0],0,9);
    garray_set(&HT_unsat[0][0],0,9);
    double fm=0;
    double fu=0;
    chyba=give_LNH(signet, suction, Temper, qstatev, hypo_L, hypo_N, H_unsat, HT_unsat, fm, fu);
    garray_move(&hypo_L[0][0][0][0], &DDtan[0][0][0][0], 81);
    convert4th_to_abaqus(DDtan, DDtan_abq);
    for(int i=0; i<6; i++) {
      for(int j=0; j<6; j++) {
         DDtan_gen[i*ngstrain+j]=multfactor*DDtan_abq[i][j];
      }
    }
  }
  //Standard update
  else if (flag == 2) {

    convert_from_general(dstrain_gen, deps, othervar, (ngstrain-6), 2);
    dsuction=othervar[0];
    dTemper=othervar[1];
    
    double dsignet[9];
    garray_set(dsignet, 0, 9);

    double Sr_new=0;
    long thischyba=calc_fsigq(signet, suction, Temper, qstatev, deps, dsuction, dTemper, dsignet, Sr_new, dqstatev);
    if(!chyba) chyba=thischyba;
    
    //Conversion of input to Abaqus convention
    double othervar[1];
    othervar[0]=Sr_new-Sr;
    garray_set(dstress_gen,0,ngstress);
    convert_to_general(dstress_gen, dsignet, othervar,(ngstress-6),1);
  }
  //Called with wrong flag 
  else {
      cout<<"subroutine calc_fsigq called with unknown flag"<<endl;
      exit(0);
  }
  
  if(!chyba) chyba = test_math_err();
  errno = 0;
  return(chyba);
};

/**
  The function calculates increment of net stress vector, of state variable vector and new Sr.

  @param signet[9] - (input): total stress: old value. This is not to be updated, increment of stress is in dsignet[9].
  @param suction - (input): suction: old value. This is not to be updated.
  @param Temper - (input): temperature: old value. This is not to be updated.
  @param qstatev[] - (input): state variable vector: old values. This is only to be updated if flag==0, otherwise only their increment is produced in dqstatev[].
  @param deps[9] - (input): increment of total strain
  @param dsuction - (input): increment of pore air pressure
  @param dTemper - (input): increment of temperature
  @param dsignet[9] - (output): increment of net stress
  @param Sr - (output): new degree of saturation
  @param dqstatev[] - (output): increment of state variables

  @return The function does not return anything.
 
  Created by D. Masin 13.7.2015
*/


long Hypoplasti_unsat_expansive_thermal::calc_fsigq(double signet[9], double suction, double Temper, 
  double *qstatev, double deps[9], double dsuction, double dTemper, 
  double dsignet[9], double &Sr_new, double dqstatev[]) {
    
  //calculation of tangent oprators L, N, Hs, HT    
  long wrongstate=0;
  long thiswrongstate=0;
  double kron_delta[3][3];
  garray_set(&kron_delta[0][0],0,9);
  kron_delta[0][0]=kron_delta[1][1]=kron_delta[2][2]=1;
  double hypo_L[3][3][3][3];
  double hypo_N[3][3];
  double H_unsat[3][3];
  double HT_unsat[3][3];
  garray_set(&hypo_L[0][0][0][0],0,81);
  garray_set(&hypo_N[0][0],0,9);
  garray_set(&H_unsat[0][0],0,9);
  garray_set(&HT_unsat[0][0],0,9);
  double fm=0.5;
  double fu=0;

  thiswrongstate=give_LNH(signet, suction, Temper, qstatev, hypo_L, hypo_N, H_unsat, HT_unsat, fm, fu);
  
  if(!wrongstate) wrongstate=thiswrongstate;
  double eM=qstatev[5];
  double SrM=qstatev[6];
  double a_scan=qstatev[7];
  double re=qstatev[8];
  if(re<0) {
    cout<<"re<0 for some reason, correcting it, just for fm calculation. re="<<re<<endl;
    re=0;
  }
    
  //Macrostructural strain increment for the given iteration
  double depM[9];
  garray_move(deps,depM,9);
  //Macrostructural stress increment for the given iteration
  double dsigMef[9];
  garray_set(dsigMef,0,9);
  //Total stress increment for the given iteration
  double dsignet_iter[9];
  garray_set(dsignet_iter,0,9);
  //Iteration counters
  int num_iter=0;
  double error_measure=1.e10;
  //Numerical parameter controlling step size for trial-and-error search for depm
  /*
  double step_depm=garray_size(&deps[0], 9);
  double minstep=1.e-5;
  double mexerror=step_depm/500;
  mexerror=round_to_digits(mexerror,1);
  if(mexerror<5.e-10) mexerror=5.e-10;
  bool tried=false;
  if(step_depm<minstep) step_depm=minstep;
  */
  
  double mexerror=garray_size(&deps[0], 9)/1000;
  mexerror=round_to_digits(mexerror,1);
  if(mexerror<1.e-10) mexerror=1.e-10;
  
  //Trial trace of dep_m for the given iteration
  double trdepm=0;

  double sewM=aer*sairentry0*eM0/eM*(aTparam+bTparam*Temper)/(aTparam+bTparam*Trparam);
  double rlambda=0;
  if(a_scan<1 && a_scan>0) rlambda=rlambdascan;
  else if((a_scan>=1 && dsuction>0) || (a_scan<=0 && dsuction<0 && suction<=sewM)) rlambda=rlambdascan;
  else if(suction<=sewM) rlambda=1;
  bool swelling=true;

  while(1) {
  //calculation of the double structure coupling factor    
    swelling=true;  
    if(trdepm<0) swelling=false;
    fm=(csh*(suction/sewM));
    if(swelling) fm=1-gscalar_power(re,mmparam); //swelling
    if(fm>1) fm=1;
    else if(fm<0) fm=0;
    if(num_iter == 0 || num_iter == 1 || num_iter == 5 || num_iter % 50 == 0) {
        thiswrongstate=give_LNH(signet, suction, Temper, qstatev, hypo_L, hypo_N, H_unsat, HT_unsat, fm, fu);
        if(!wrongstate) wrongstate=thiswrongstate;
    }
    
    depM[0]=deps[0]-fm*trdepm/3;
    depM[4]=deps[4]-fm*trdepm/3;
    depM[8]=deps[8]-fm*trdepm/3;
    
    double DM[3][3];
    garray_set(&DM[0][0], 0, 3*3); 
    garray_move(depM, &DM[0][0], 3*3);
    double normDM=garray_size(&DM[0][0], 3*3);
        
    double linear[3][3];
    garray_set(&linear[0][0], 0, 3*3); 
    gmatrix_a4b(hypo_L, *DM, *linear);
    
  //Calculate effective stress rate of Macrostructure
    for(int i=0; i<3; i++) {
      for(int j=0; j<3; j++) {
        dsigMef[3*i+j]= linear[i][j] + hypo_N[i][j] * normDM;	
        if(dsuction>0) dsigMef[3*i+j]+=rlambda*H_unsat[i][j]*dsuction;
        if(dTemper>0) dsigMef[3*i+j]+=HT_unsat[i][j]*dTemper;
      }
    }

  //Calculate total stress rate from the effective stress rate of Macrostructure
    double corr_s=(1-lambdap0*rlambda)*dsuction;
    double deM=(1+eM)*(gtrace(DM)+(fm-1)*trdepm);
    double corr_eM=-lambdap0*suction*deM/eM;
    double corr_T=+lambdap0*suction*bTparam*dTemper/(aTparam+bTparam*Temper);
    if(suction>sewM) corr_eM=corr_T=0;
     
    for(int i=0; i<3; i++) {
      for(int j=0; j<3; j++) {
         dsignet_iter[3*i+j]=dsigMef[3*i+j]-kron_delta[i][j]*SrM*(corr_s+corr_eM+corr_T);
      }
    }
    
    if(eM<0.0001) {//for zero eM all the deformation is due to microstructure
      fu=0;
      garray_set(depM,0,9);
      fm=1;
      trdepm=(deps[0]+deps[4]+deps[8]);
      double pefsat=(signet[0]+signet[4]+signet[8])/3+suction;
      double dpefsat=-(trdepm-alpha_s*dTemper)*pefsat/kappam;
      dsignet_iter[0]=dsignet_iter[4]=dsignet_iter[8]=(dpefsat-dsuction);
    }
    
  //Calculate microstructural strain corresponding to the obtained dsignet in the current iteration.
    double depm_iter[9];
    garray_set(depm_iter,0,9);
    calc_depm(signet, suction, Temper, dsuction, dTemper, dsignet_iter, depm_iter, fu);
     
  //Calculate the difference between microstructural strain obtained in the current iteration and the trial strain trdepm
    //double old_error_measure=error_measure;
    error_measure=trdepm-3*depm_iter[0];
    
  //Solution found   
    if(gscalar_dabs(error_measure)<mexerror) break;
 
  //Solution not found, update trdepm
    /*
    if(gscalar_dabs(error_measure)>gscalar_dabs(old_error_measure)) {
        if((error_measure<0 && old_error_measure<0) || (error_measure>0 && old_error_measure>0))
            if(num_iter!=0) step_depm*=-0.5;
        else step_depm*=-1;
    }
    else if((old_error_measure>0 && error_measure<0) || (old_error_measure<0 && error_measure>0)) {
        if(num_iter!=0) step_depm*=-0.5;
    }
    double minsteprun=1.e-10;
    if(step_depm>0 && step_depm<minsteprun) step_depm=minsteprun;
    else if(step_depm<0 && step_depm>-1*minsteprun) step_depm=-1*minsteprun;
    if(num_iter>800 && !tried) {
         cout<<"Microstructural strain: critical number of iterations, err="<<error_measure<<endl;
         step_depm*=1.e6;
         tried=true;
    }
    trdepm+=step_depm;
    */
    
    trdepm=(3*depm_iter[0]+3*trdepm)/4;
    
    if(num_iter>1000) {
       bool debug=false;
       cout<<"Microstructural strain: maximum number of iterations, err="<<error_measure<<endl;
       if(debug) {
           cout<<"trdepm="<<trdepm<<endl;
           cout<<"tr(depm_iter)="<<3*depm_iter[0]<<endl;
           cout<<"signet[0], signet[1], signet[2], signet[4], signet[5], signet[8]="<<signet[0]<<" "<<signet[1]<<" "<<signet[2]<<" "<<signet[4]<<" "<<signet[5]<<" "<<signet[8]<<endl;
           cout<<"dsignet_iter[0], dsignet_iter[4], dsignet_iter[8]="<<dsignet_iter[0]<<" "<<dsignet_iter[4]<<" "<<dsignet_iter[8]<<endl;
           cout<<"fm="<<fm<<endl;
           cout<<"depM[0], depM[4], depM[8]="<<depM[0]<<" "<<depM[4]<<" "<<depM[8]<<endl;
           cout<<"deps[0], deps[1], deps[2], deps[4], deps[5], deps[8]="<<deps[0]<<" "<<deps[1]<<" "<<deps[2]<<" "<<deps[4]<<" "<<deps[5]<<" "<<deps[8]<<endl;
           cout<<"e, eM, SrM, a_scan, re="<<qstatev[0]<<" "<<qstatev[5]<<" "<<qstatev[6]<<" "<<qstatev[7]<<" "<<qstatev[8]<<endl;
       }
       return(1);
    }
    num_iter++;
//        cout<<gscalar_dabs(error_measure)<<endl;
  }
//     cout<<" end iter "<<num_iter<<endl;
  garray_move(dsignet_iter, dsignet, 9); 
    
  //this calculates dqstatev, it is not updating the original values
  int flag=2;
  
  thiswrongstate=update_hisv(signet, suction, Temper, qstatev, deps, dsuction, dTemper, dsignet, Sr_new, dqstatev, flag, fu);
  if(!wrongstate) wrongstate=thiswrongstate;
  
  if(wrongstate) {cout<<"Error in THM hypoplastic model, step decrease requested"<<endl;}
  return(wrongstate);
}

/**
  The function calculates state variable increment vector dqstatev[]. The old values of qstatev[0] (void ratio) and qstatev[6] (a_scan) must be provided by the FEM (even for the initialisation stage). The order of state variables is as follows: 
  [0]e (required on input)
  [1]suction
  [2]Sr
  [3]T
  [4]em
  [5]eM
  [6]SrM
  [7]a_scan (required on input)
  [8]re

  @param signet[9] - (input): total stress: old value. This is not to be updated, increment of stress is in dsignet[9].
  @param suction - (input): suction: old value. This is not to be updated.
  @param Temper - (input): temperature: old value. This is not to be updated.
  @param qstatev[] - (input): state variable vector: old values. This is only to be updated if flag==0, otherwise only their increment is produced in dqstatev[].
  @param deps[9] - (input): increment of total strain
  @param dsuction - (input): increment of pore air pressure
  @param dTemper - (input): increment of temperature
  @param dsignet[9] - (output): increment of total stress
  @param Sr - (output): new degree of saturation
  @param dqstatev[] - (output): increment of state variables

  @return The function does not return anything.
 
  Created by D. Masin 13.7.2015
*/

long Hypoplasti_unsat_expansive_thermal::update_hisv(double signet[9], double suction, double Temper, double qstatev[], 
  double deps[9], double dsuction, double dTemper, 
  double dsignet[9], double &Sr, double dqstatev[], int flag, double fu) {

  double D[3][3];
  garray_set(&D[0][0], 0, 3*3);
  garray_move(deps, &D[0][0], 3*3);

  //old values of state variables needed for calculations
  double evoid=qstatev[0];
  double a_scan=qstatev[7];

  //calculating state variable increments/new values
  double delta_e=(1+evoid)*gtrace(D); 
  double evoid_new=evoid+delta_e;
  double Temper_new=Temper+dTemper;
  
  double signet_new[9];
  garray_add(signet, dsignet, signet_new, 9);
  double suction_new=suction+dsuction;
  double emstarT=exp(alpha_s*(Temper-Trparam)+log(1+emstar))-1;
  double emstarT_new=exp(alpha_s*(Temper_new-Trparam)+log(1+emstar))-1;

  double pefsat_new=(signet_new[0]+signet_new[4]+signet_new[8])/3+suction_new;
  double pefsat=(signet[0]+signet[4]+signet[8])/3+suction;
  double em=qstatev[4];
  if(fu>1) fu=1;
  double d_em=(1+em)*(alpha_s*dTemper-kappam/pefsat*(pefsat_new-pefsat));
  if(d_em>0) d_em*=(1-fu);
  double em_new=em+d_em;
  
  if(flag==0) {
      em_new=exp(kappam*log(smstar/pefsat_new)+log(1+emstarT_new))-1;
      em=exp(kappam*log(smstar/pefsat)+log(1+emstarT))-1;      
  }
  double eM=(evoid-em)/(1+em);
  double eM_new=(evoid_new-em_new)/(1+em_new);
  
  if(smstar>=0) {
    cout<<"smstar must be lower than 0, returning error information"<<endl;
    return(1);
  }
  if(eM_new<-0.1) {
    cout<<"eM<0, this is not good, returning error information, eM="<<eM_new<<endl;
    return(1);
  }
  if(em_new<0) {
    cout<<"em<0, this is not good, returning error information"<<endl;
    return(1);
  }
  
  if (a_scan>1.001 || a_scan<-0.001) {
    cout<<"something wrong with a_scan, a_scan=" << a_scan << endl;
    return(1);	
  }
  double eMuse=eM;
  if(eM<0.01) eMuse=0.01;
  double eM_newuse=eM_new;
  if(eM<0.01) eM_newuse=0.01;
  
  double sedM=sairentry0*eM0/eMuse*(aTparam+bTparam*Temper)/(aTparam+bTparam*Trparam);
  double sewM=sedM*aer;
  double sedM_new=sairentry0*eM0/eM_newuse*(aTparam+bTparam*Temper_new)/(aTparam+bTparam*Trparam);

  double se=sedM*(aer+a_scan-aer*a_scan); 
  
  double a_scan_new=0;
  double sD=0.0;
  if (suction<sewM) {
    sD=sedM/se*suction;
    double d_a_scan=(1-rlambdascan)/(sD*(1.0-aer))*(suction_new-suction);
    a_scan_new=a_scan+d_a_scan;
    if (a_scan_new>=1.0) a_scan_new=1.0;
    if (a_scan_new<=0.0) a_scan_new=0.0;
  }
  double se_new=sedM_new*(aer+a_scan_new-aer*a_scan_new); 
  double SrM_new=1;
  if(suction_new<sewM) SrM_new=gscalar_power((se_new/suction_new), lambdap0);

  double Sr_new=(SrM_new*(evoid_new-em_new)+em_new)/evoid_new;
  Sr=Sr_new;

  double tmpscal=0;
  double Nuse_new=0;
  double lambda_use_new=0;
  
  //fill-in state variable increments  
  dqstatev[0]=evoid_new-qstatev[0];
  dqstatev[1]=suction_new-qstatev[1];
  dqstatev[2]=Sr_new-qstatev[2];
  dqstatev[3]=Temper_new-qstatev[3];
  dqstatev[4]=em_new-qstatev[4];
  dqstatev[5]=eM_new-qstatev[5];
  dqstatev[6]=SrM_new-qstatev[6];
  dqstatev[7]=a_scan_new-qstatev[7];
  dqstatev[8]=0; //re: to be filled later
  
  //calculate new state variables, needed just for re
  double qstatev_new[nstatev];
  garray_add(qstatev, dqstatev, qstatev_new, nstatev);  
  
  //calculate new re
  calc_N_lambda_dpeds(Nuse_new, lambda_use_new, tmpscal, tmpscal, signet_new, suction_new, Temper_new, qstatev_new);
  double sigef_new[9];
  make_sig_unsat(signet_new, suction_new, SrM_new, sigef_new, suction_new);
  double pmeanef_new=-(sigef_new[0]+sigef_new[4]+sigef_new[8])/3;
  double emaxisot_new=exp(Nuse_new-lambda_use_new*1)-1;
  if(pmeanef_new>1) emaxisot_new=exp(Nuse_new-lambda_use_new*log(pmeanef_new))-1;
  double re_new=0.0001;
  if(emaxisot_new>em_new) re_new=(evoid_new-em_new)/(emaxisot_new-em_new);
  if(re_new<0) re_new=0.0001;
  double re=qstatev[8];
  double dre=re_new-re;
  dqstatev[8]=dre;
  return(0);
}

/**
  The function calculates the mascrostructural effective stress.

  @param sig[9] - (input): net stress.
  @param suction - (input): pore air pressure.
  @param SrM - (input): degree of saturation of macrostructure.
  @param tensor_unsat[9] - (output): effective stress of macrostructure
  @param scalar_unsat - (output): suction

  @return The function does not return anything.
 
  Created by D. Masin 13.7.2015
*/

void Hypoplasti_unsat_expansive_thermal::make_sig_unsat(double sig[3*3], double suction, double SrM,
  double tensor_unsat[3*3], double &scalar_unsat) {

  garray_set(tensor_unsat, 0, 3*3);
  garray_move(sig, tensor_unsat, 3*3);
  scalar_unsat=suction;

  double chi=SrM;

  tensor_unsat[0]+=chi*scalar_unsat;
  tensor_unsat[4]+=chi*scalar_unsat;
  tensor_unsat[8]+=chi*scalar_unsat;
}

/**
  The function calculates the microstructural strain.

  @param signet[9] - (input): total stress.
  @param suction - (input): suction
  @param Temper - (input): temperature.
  @param dsuction - (input): increment of matric suction
  @param dTemper - (input): increment of temperature
  @param dsignet[9] - (input!!!): increment of total stress
  @param depm[9] - (output): microstructural strain

  @return The function does not return anything.
 
  Created by D. Masin 13.7.2015
*/

void Hypoplasti_unsat_expansive_thermal::calc_depm(double signet[9], double suction, double Temper, double dsuction, 
  double dTemper, double dsignet[9], double depm[9], double fu) {
    
  double pefsat=(signet[0]+signet[4]+signet[8])/3+suction;
  double dpefsat=(dsignet[0]+dsignet[4]+dsignet[8])/3+dsuction;
        
  garray_set(depm,0,9);
  depm[0]=depm[4]=depm[8]=-kappam*(dpefsat/pefsat)/3;
  for(int i=0; i<3; i++) depm[3*i+i]+=alpha_s*dTemper/3;
  if(fu>1) fu=1;
  if((depm[0]+depm[4]+depm[8])>0) {
    for(int i=0; i<3; i++) depm[3*i+i]*=(1-fu);
  }
}

/**
  The function calculates current position and slope of normal compression line Nuse and lambda_use, and also the derivatives (dpe/ds)/pe and (dpe/dT)/pe

  @param Nuse - (output): Nuse
  @param lambda_use - (output): lambda_use
  @param dpeds_divpe - (output): (dpe/ds)/pe
  @param dpedT_divpe - (output): (dpe/dT)/pe
  @param signet[9] - (input): total stress.
  @param suction - (input): suction.
  @param Temper - (input): temperature.
  @param qstatev[] - (input): state variable vector.

  @return The function does not return anything.
 
  Created by D. Masin 13.7.2015
*/

void Hypoplasti_unsat_expansive_thermal::calc_N_lambda_dpeds(double &Nuse, double &lambda_use, double &dpeds_divpe, double &dpedT_divpe, double signet[9], double suction, double Temper, double qstatev[]) {
    
  double evoid=qstatev[0];
  double SrM=qstatev[6];

  double logsse=log(1/(gscalar_power(SrM,(1/lambdap0))));

  double sig_ef[9];
  make_sig_unsat(signet, suction, SrM, sig_ef, suction);

  Nuse=N+nparam*logsse+nTparam*log(Temper/Trparam);
  lambda_use=lambda+lparam*logsse+lTparam*log(Temper/Trparam);

  double pe=exp((Nuse-log(1+evoid))/lambda_use);
  dpeds_divpe=0;
  
  if(SrM<1) dpeds_divpe=(nparam-log(pe)*lparam)/(lambda_use*suction);
  dpedT_divpe=(nTparam-log(pe)*lTparam)/(lambda_use*Temper);
//  if(dpeds_divpe>0) {
//      cout<<"Check lparam and lparam, dpeds<0!!"<<endl;
//      exit(0);
//     dpeds_divpe=0;
//  }
//  if(dpedT_divpe>0) {
//    cout<<"Check lTparam and lTparam, dpedT>0!!"<<endl;
//    exit(0);
//  }
};

/**
  The function calculates tangent operators hypo_L, hypo_N, H_unsat and HT_unsat of the model for macrostructure

  @param signet[9] - (input): total stress.
  @param suction - (input): suction.
  @param Temper - (input): temperature.
  @param qstatev[] - (input): state variable vector.
  @param hypo_L[3][3][3][3] - (output): tangent operator hypo_L
  @param hypo_N[3][3] - (output): tangent operator hypo_N
  @param H_unsat[3][3] - (output): tangent operator H_unsat
  @param HT_unsat[3][3] - (output): tangent operator HT_unsat

  @return The function does not return anything.
 
  Created by D. Masin 13.7.2015
*/

long Hypoplasti_unsat_expansive_thermal::give_LNH(double signet[9], double suction, 
  double Temper, double qstatev[], double hypo_L[3][3][3][3], double hypo_N[3][3], double H_unsat[3][3], double HT_unsat[3][3], double fm, double &fu) { 

  long ret=0;
  double sig_ef[9];
  double stresssuction=0;
  double SrM=qstatev[6];
  make_sig_unsat(signet, suction, SrM, sig_ef, stresssuction);
  double T[3][3];
  garray_set(&T[0][0], 0, 3*3); 
  garray_move(sig_ef, &T[0][0], 3*3);
  double p=-gtrace(T)/3;
  if(p<0) {
    cout<<"p<0, this is error, ending calculation p="<<p<<" suction="<<suction<<" pnet="<<signet[0]+signet[4]+signet[8]<<endl;
    exit(0);
  }
  
  double kron_delta[3][3];
  garray_set(&kron_delta[0][0],0,9);
  kron_delta[0][0]=kron_delta[1][1]=kron_delta[2][2]=1;

  double T_star[3][3];
  garray_move( &T[0][0], &T_star[0][0], 3*3 );				
  for ( int idim=0; idim<3; idim++ ) T_star[idim][idim] += p;
  double T_str_star[3][3];
  garray_multiply(&T_star[0][0], &T_str_star[0][0], (1/gtrace(T)), 3*3);
  double T_str[3][3];
  garray_multiply(&T[0][0], &T_str[0][0], (1/gtrace(T)), 3*3);
  double e=qstatev[0];
  double I1=gtrace(T);
  double I2=(garray_inproduct(&T[0][0], &T[0][0], 3*3)-I1*I1)/2;
  double I3=gmatrix_determinant(&T[0][0], 3);

  double Nuse=N;
  double lambda_use=lambda;
  double dpeds_divpe=0;
  double dpedT_divpe=0;
  calc_N_lambda_dpeds(Nuse, lambda_use, dpeds_divpe, dpedT_divpe, signet, suction, Temper, qstatev);

  double peast=exp((Nuse-log(1+e))/lambda_use);	
//        cout<<"OCR="<<peast/p<<endl;
  double fd=(2*p)/peast;
  double cos2phic=1-sin(phic)*sin(phic);
  double sin2phim=(9*I3+I1*I2)/(I3+I1*I2);
  bool isotrop=false;
  if(sin2phim<1.e-10) {
    sin2phim=0;
    isotrop=true;
  }
  if(sin2phim>(1-1.e-10)) {
      //ret=2; //this was uncommented but the error is still there
      sin2phim=(1-1.e-10);
  }

  double ashape=0.3;
  double npow=-log(cos2phic)/log(2)+ashape*(sin2phim-sin(phic)*sin(phic));
  double phimfactor=-8*I3/(I3+I1*I2);
  if(phimfactor<1.e-10) phimfactor=1.e-10;
  double fdsbs=2*gscalar_power(phimfactor,(1/npow));
  double a=sqrt(3.)*(3-sin(phic))/(2*sqrt(2.)*sin(phic));
  double alpha_power=log((lambda-kappa)*(3+a*a)/((lambda+kappa)*a*sqrt(3.)))/log(2.); 
  fd=gscalar_power(fd,alpha_power);
  fdsbs=gscalar_power(fdsbs,alpha_power);
  if(fdsbs<fd && sin2phim>(1-1.e-9)) fdsbs=fd;
	
  double fddivfdA=fd/fdsbs;
  double nuhh=nuparam;
  double nuvh=nuhh/alphanu;

  double Am=nuvh*nuvh*(4*alphaE*alphanu-2*alphaE*alphaE*alphanu*alphanu+2*alphaE*alphaE-alphanu*alphanu)+nuvh*(4*alphaE+2*alphaE*alphanu)+1+2*alphaE;
  double fs=9*p/2*(1/kappa+1/lambda_use)/Am;

  double a1=alphaE*(1-alphanu*nuvh-2*alphaE*nuvh*nuvh);
  double a2=alphaE*nuvh*(alphanu+alphaE*nuvh);
  double a3=alphaE*nuvh*(1+alphanu*nuvh-alphanu-alphaE*nuvh);
  double a4=(1-alphanu*nuvh-2*alphaE*nuvh*nuvh)*(alphaE*(1-alphaG))/alphaG;
  double a5=alphaE*(1-alphaE*nuvh*nuvh)+1-alphanu*alphanu*nuvh*nuvh-2*alphaE*nuvh*(1+alphanu*nuvh)-2*alphaE*(1-alphanu*nuvh-2*alphaE*nuvh*nuvh)/alphaG;

  garray_set(&hypo_L[0][0][0][0], 0, 3*3*3*3);
  double nvect[3];
  garray_set(nvect,0,3);
  nvect[0]=1;//vertical direction is "0"
  double pmat[3][3];
  garray_set(&pmat[0][0],0,9);
  for(int i=0; i<3; i++) {
    for(int j=0; j<3; j++) {		
      pmat[i][j]=nvect[i]*nvect[j];
    }
  }
  double kck[3][3][3][3];
  double kdk[3][3][3][3];
  double pdk[3][3][3][3];
  double kdp[3][3][3][3];
  double pck[3][3][3][3];
  double pdp[3][3][3][3];
  garray_set(&kck[0][0][0][0], 0, 81);	
  garray_set(&kdk[0][0][0][0], 0, 81);	
  garray_set(&pdk[0][0][0][0], 0, 81);	
  garray_set(&kdp[0][0][0][0], 0, 81);	
  garray_set(&pck[0][0][0][0], 0, 81);	
  garray_set(&pdp[0][0][0][0], 0, 81);	
  for ( int i=0; i<3; i++ ) {
    for ( int j=0; j<3; j++ ) {
      for ( int k=0; k<3; k++ ) {
        for ( int l=0; l<3; l++ ) {
          kck[i][j][k][l]=(kron_delta[i][k]*kron_delta[j][l]+kron_delta[i][l]*kron_delta[j][k]+kron_delta[j][l]*kron_delta[i][k]+kron_delta[j][k]*kron_delta[i][l])/2;
          kdk[i][j][k][l]=kron_delta[i][j]*kron_delta[k][l];
          pdk[i][j][k][l]=pmat[i][j]*kron_delta[k][l];
          kdp[i][j][k][l]=kron_delta[i][j]*pmat[k][l];
          pck[i][j][k][l]=(pmat[i][k]*kron_delta[j][l]+pmat[i][l]*kron_delta[j][k]+pmat[j][l]*kron_delta[i][k]+pmat[j][k]*kron_delta[i][l])/2;
          pdp[i][j][k][l]=pmat[i][j]*pmat[k][l];
        }
      }
    }
  }
  for ( int i=0; i<3; i++ ) {
    for ( int j=0; j<3; j++ ) {
       for ( int k=0; k<3; k++ ) {
          for ( int l=0; l<3; l++ ) {
            hypo_L[i][j][k][l]=a1*kck[i][j][k][l]/2+a2*kdk[i][j][k][l]+a3*(pdk[i][j][k][l]+kdp[i][j][k][l])+a4*pck[i][j][k][l]+a5*pdp[i][j][k][l];	
          }
       }
    }
  }

  bool unsat_active=false;
  if(SrM<1) unsat_active=true;
  
  //get lambda_act for constant suction;
  double eM=qstatev[5];
  double em=qstatev[4];
  double pefsat=-((signet[0]+signet[4]+signet[8])/3+suction);
  double lambda_act=lambda_use;
  if (unsat_active) {
//      lambda_act=lambda_use/(1.0-(nparam-lparam*log(p))*(1.0+e)/e);
        lambda_act=(lambda_use*eM*(1+em)-(p/pefsat)*kappam*(1+eM)*(1+em)*(nparam-lparam*log(p))-(p/pefsat)*kappam*fm*eM*(1+em))/(eM*(1+em)-(1+e)*(nparam-lparam*log(p)));
//      double A=kappam*(1+eM)*(nparam-lparam*log(p))/eM+kappam*fm;
//      double citatel=lambda_use-A*(p/pefsat)*(1-kappam*(1+eM)*SrM*lambdap0*suction/(eM*pefsat));
//      double jmenovatel=1-(1+e)/(eM*(1+em))*(nparam-lparam*log(p)-A*SrM*lambdap0*suction/pefsat);
//      lambda_act=citatel/jmenovatel;
  }
//     cout<<signet[0]<<" "<<-suction<<" "<<Temper<<" "<<p/peast<<endl;
  fs=9*p/2*(1/kappa+1/lambda_act)/Am;
  garray_multiply(&hypo_L[0][0][0][0], &hypo_L[0][0][0][0], fs, 81);
    
  double hypo_Dsom[3][3];
  garray_set(&hypo_Dsom[0][0], 0, 3*3);
  double kpow=1.7+3.9*sin(phic)*sin(phic);
  double sinphickpow=gscalar_power(sin(phic),kpow);
  double sinphimkpow=gscalar_power(sqrt(sin2phim),kpow);

  double TT[3][3];
  garray_set(&TT[0][0], 0, 3*3);
  gmatrix_ab(&T_str_star[0][0], &T_str_star[0][0], &TT[0][0], 3, 3, 3);
  double TTT[3][3];
  garray_set(&TTT[0][0], 0, 3*3);
  gmatrix_ab( &TT[0][0], &T_str_star[0][0], &TTT[0][0],3, 3, 3);
  double cos3theta=-sqrt(6.)*gtrace(TTT)/gscalar_power(gtrace(TT),1.5);
  if(isotrop) cos3theta=-1;

  double Amult=2./3-sqrt(sqrt(sin2phim))*(cos3theta+1)/4.;

  for(int i=0; i<3; i++) {
    for(int j=0; j<3; j++) {		
      hypo_Dsom[i][j]=-T_str_star[i][j]+kron_delta[i][j]*(sinphimkpow-sinphickpow)/(1-sinphickpow)*Amult;
    }
  }
  double Dsomnorm = garray_size(&hypo_Dsom[0][0], 3*3);
  garray_multiply(&hypo_Dsom[0][0], &hypo_Dsom[0][0], (1/Dsomnorm), 3*3);

  double A[3][3][3][3];
  garray_set(&A[0][0][0][0], 0, 81);
        
  for ( int i=0; i<3; i++ ) {
    for ( int j=0; j<3; j++ ) {
      for ( int k=0; k<3; k++ ) {
        for ( int l=0; l<3; l++ ) {
          A[i][j][k][l]=hypo_L[i][j][k][l]+T[i][j]*kron_delta[k][l]/lambda_act;
        }
      }
    }
  }
        
  gmatrix_a4b(A, &hypo_Dsom[0][0], &hypo_N[0][0]);
  garray_multiply(&hypo_N[0][0], &hypo_N[0][0],-fddivfdA, 3*3);

  double pdivpsbs=gscalar_power(fddivfdA, (1/alpha_power));
  fu=gscalar_power(pdivpsbs, mparam);
  if(pdivpsbs>1 && pdivpsbs<5) fu=gscalar_power(pdivpsbs, 10);
  else if (pdivpsbs>5) fu=gscalar_power(5, 10);
  double cic=1;

  bool m0_OCRindep_collapse=true;
  if(m0_OCRindep_collapse && fd<fdsbs) {
    cic=(lambda_act+kappa)*(gscalar_power(2.0,alpha_power)-fd)+2*kappa*fd;
    cic=cic/((lambda_act+kappa)*(gscalar_power(2.0,alpha_power)-fdsbs)+2*kappa*fdsbs);	
  }

  garray_set(&H_unsat[0][0], 0, 3*3);
  garray_multiply(&T[0][0], &H_unsat[0][0], dpeds_divpe*cic*lambda_use/lambda_act, 9);	
  garray_set(&HT_unsat[0][0], 0, 3*3);
  garray_multiply(&T[0][0], &HT_unsat[0][0], dpedT_divpe*cic*lambda_use/lambda_act, 9);
  garray_multiply(&H_unsat[0][0], &H_unsat[0][0], fu, 9);  //H times fd/fdSOM
  garray_multiply(&HT_unsat[0][0], &HT_unsat[0][0], fu, 9);  //HT times fd/fdSOM
  return(ret);
}

//Functions for abaqus-full convention conversion

void General_model::convert_from_general( double general[], double stressstrain[], double othervar[], int nothv, int flag ) {
//flag==1: stress-like; flag==2: strain-like
//4 -> 12 , 5 -> 13, 6 -> 23
  stressstrain[0]=general[0];
  stressstrain[4]=general[1];
  stressstrain[8]=general[2];
  if(flag == 1) {
    stressstrain[1]=general[3];
    stressstrain[2]=general[4];
    stressstrain[5]=general[5];
    stressstrain[3]=general[3];
    stressstrain[6]=general[4];
    stressstrain[7]=general[5];
  }
  if(flag == 2) {
    stressstrain[1]=general[3]/2;
    stressstrain[2]=general[4]/2;
    stressstrain[5]=general[5]/2;
    stressstrain[3]=general[3]/2;
    stressstrain[6]=general[4]/2;
    stressstrain[7]=general[5]/2;
  }    
  for(int i=0; i<nothv; i++) othervar[i]=general[6+i];
}

void General_model::convert_to_general( double general[], double stressstrain[], double othervar[], int nothv, int flag ){
//flag==1: stress-like; flag==2: strain-like
//4 -> 12 , 5 -> 13, 6 -> 23
  general[0]=stressstrain[0];
  general[1]=stressstrain[4];
  general[2]=stressstrain[8];
  if(flag == 1) {
    general[3]=stressstrain[1];
    general[4]=stressstrain[2];
    general[5]=stressstrain[5];
  }
  if(flag == 2) {
    general[3]=2*stressstrain[1];
    general[4]=2*stressstrain[2];
    general[5]=2*stressstrain[5];
  }   
  for(int i=0; i<nothv; i++) general[6+i]=othervar[i];
}

void General_model::convert4th_to_abaqus( double DDfull[3][3][3][3], double DDabq[6][6]){
  int fi=0;
  int fj=0;
  int fk=0;
  int fl=0;
  for ( int ai=0; ai<6; ai++ ) {
    for ( int aj=0; aj<6; aj++ ) {
        map_indices_abqtofull( fi, fj, ai);
        map_indices_abqtofull( fk, fl, aj);
        DDabq[ai][aj]=(DDfull[fi][fj][fk][fl]+DDfull[fi][fj][fl][fk])/2;
    }
  }
}
void General_model::map_indices_fulltoabq( int fulli, int fullj, int &abq){
   if(fulli == 0 && fullj == 0) abq=0;
   if(fulli == 1 && fullj == 1) abq=1;
   if(fulli == 2 && fullj == 2) abq=2;
   if(fulli == 0 && fullj == 1) abq=3;
   if(fulli == 0 && fullj == 2) abq=4;
   if(fulli == 1 && fullj == 2) abq=5;
   if(fulli == 1 && fullj == 0) abq=3;
   if(fulli == 2 && fullj == 0) abq=4;
   if(fulli == 2 && fullj == 1) abq=5;
}
void General_model::map_indices_abqtofull( int &fulli, int &fullj, int abq){
   if(abq==0) {
     fulli=0;
     fullj=0;
   };
   if(abq==1) {
     fulli=1;
     fullj=1;
   };
   if(abq==2) {
     fulli=2;
     fullj=2;
   };
   if(abq==3) {
     fulli=0;
     fullj=1;
   };
   if(abq==4) {
     fulli=0;
     fullj=2;
   };
   if(abq==5) {
     fulli=1;
     fullj=2;
   };
}

//Functions for mathematical oprations

double General_model::gscalar_power( double a, double b ) {
  /*  if(b==0.5 && a<0) {
	cout<<"Attempting to calculate square root of a negative number";
	exit(0);
  }		
  double result=0.;
  if ( b==0. )
    result = 1.;
  else
  result = pow( a, b );
  return result;*/
  double ret = pow(a,b);
  if (test_math_err() == 0)
    return ret;
  else
  {
    cout<<"Attempting to calculate square root of a negative number";
    exit(0);
  }  
  return 0.0;
}

double General_model::gscalar_dabs(double k) {
	double tmp;
	if(k>= 0) tmp=k;
	if(k<0) tmp=-1*k;
	return tmp;
}

void General_model::garray_add( double a[], double b[], double c[], long int n ) {
  register long int i=0;
  for ( i=0; i<n; i++ ) c[i] = a[i] + b[i];
}

double General_model::garray_inproduct( double a[], double b[], long int n ) {
  register long int i=0;
  double result=0.;
  for ( i=0; i<n; i++ ) result += a[i]*b[i];
  return result;
}

void General_model::garray_move( double from[], double to[], long int n ) {
  //  register long int i=0;
  //  for ( i=0; i<n; i++ ) to[i] = from[i];
  memcpy(to, from, sizeof(*from)*n);
}

void General_model::garray_multiply( double a[], double b[], double c, long int n ) {
  register long int i=0;
  for ( i=0; i<n; i++ ) b[i] = c * a[i];
}

void General_model::garray_set( double *ptr, double value, long int n ) {
  // register long int i=0;
  // for ( i=0; i<n; i++ ) *(ptr+i) = value;
  if (value == 0.0)
    memset(ptr, 0, sizeof(*ptr)*n);
  else
    for (long int i=0; i<n; i++ ) *(ptr+i) = value;
}

double General_model::garray_size( double a[], long int n ) {
  double size=0.;
  size = sqrt( gscalar_dabs( garray_inproduct(a,a,n) ) );
  return size;
}

double General_model::gtrace( double a[3][3]) {
	return(a[0][0]+a[1][1]+a[2][2]);
}

double General_model::gmatrix_determinant( double a[], long int n ) {

  double result=0.;

  if ( n==1 )
    result = a[0];
  else if ( n==2 )
    result = a[0]*a[3] - a[1]*a[2];
  else {
    if(n == 3) {
      result = a[0]*(a[4]*a[8]-a[7]*a[5]) -
      a[1]*(a[3]*a[8]-a[6]*a[5]) + a[2]*(a[3]*a[7]-a[6]*a[4]);
    }
    else {
      cout<<"n must be 1-3"<<endl;   
      exit(0);
    }
  }
  return result;
}

void General_model::gmatrix_ab( double *a, double *b, double *c, long int n, long int m,
  long int k )
  // c[n][k] = a[n][m] * b[m][k]
{
    register long int i=0, j=0, l=0;

  for ( i=0; i<n; i++ ) {
    for ( j=0; j<k; j++ ) {
      *( c + i*k + j ) = 0;
      for ( l=0; l<m; l++ ) {
        *( c + i*k + j ) += ( *( a + i*m + l ) ) *
                             ( *( b + l*k + j ) );
       }
    }
  }
}

void General_model::gmatrix_a4b( double a[3][3][3][3], double b[], double c[] )

{
  register long int i=0, j=0, k=0, l=0;

  for ( i=0; i<3; i++ ) {
    for ( j=0; j<3; j++ ) {
      c[i*3+j] = 0.;
      for ( k=0; k<3; k++ ) {
        for ( l=0; l<3; l++ ) {
          c[i*3+j] += a[i][j][k][l] * b[k*3+l];
        }
      }
    }
  }

}

double General_model::round_to_digits(double value, int digits)
{
    if (value == 0.0) // otherwise it will return 'nan' due to the log10() of zero
        return 0.0;

    double factor = pow(10.0, digits - ceil(log10(fabs(value))));
    return round(value * factor) / factor;   
}

long General_model::test_math_err()
{
  if ((errno == EDOM) || (errno == ERANGE))
    return 1;

  return 0;
}
