#include <string.h>
#include "nonlinman.h"



nonlinman::nonlinman (void)
{

  //  type of solver of nonlinear algebraic equation system
  tnlinsol = (nonlinsolvertype) 0;
  //  norm measure of displacement increments
  displnorm = (displacementnorm) 0;

  nial = 0;
  niilal = 0;
  erral = 0.0;
  dlal = 0.0;
  dlmaxal = 0.0;
  psial = 0.0;
  nsnal=0;
  selnodal=NULL;
  hdbackupal=0;
  nsdofal=0;
  seldofal=NULL;
  //  determination of increment of load parameter
  dlam=nodetermination;

  check_tot_al=off;

  ninr = 0;
  niilnr = 0;
  errnr = 0.0;
  incrnr = 0.0;
  minincrnr = 0.0;

  nienr = 0;
  hdbr = 0;
  hdbid = 0;
  backupfname[0] = 0;

  //  type of stiffness matrix
  stmat = initial_stiff;
  //  number of steps with constant stiffness %matrix
  ithstiffchange=1;
  jthstiffchange=1;

  check_div = off;
  div_min_steps = 3;
  divc_step = NULL;
  divc_err  = NULL;
  check_lambdar = off;
  lambdar = 0.0;
  errl = 0.0;
}



nonlinman::~nonlinman (void)
{
  delete [] selnodal;
  delete [] seldofal;
  delete [] divc_step;
  delete [] divc_err;
}



/**
  The function reads parameters for individual types of solvers.
  
  @param in - pointer to the opened text file
  @param mespr - flag for message printing control (0 = no messages are printed, 1 = messages are sentto stdout)

  @return The function does not return anything.

  Created by Jaroslav Kruis, 2008
  Modified by Tomas Koudelka, 08.2011
*/
void nonlinman::read (XFILE *in,long mespr)
{
  //  type of nonlinear solver
  //  1- arc-length
  //  2 - newton method
  //  3 - arclrv1
  //  4 - newtonrv1
  //  5 - newtonrestart
  //  6 - displctrl
  //  7 - displctrlrv
  //  8 - arclrv
  //  10 - adaptram
  //  20 - arclg
  //  30 - newtong
  xfscanf (in, "%k%m", "type_of_nonlin_solver", &nonlinsolvertype_kwdset, (int*)&tnlinsol);

  //  type of stiffness matrix (initial, tangent)
  //  1 - initial_stiff
  //  2 - tangent_stiff
  //  3 - secant_stiff
  //  5 - incr_tangent_stiff
  //  6 - ijth_tangent_stiff
  xfscanf (in,"%k%m","stiffmat_type", &stiffmatrix_kwdset, (int*)&stmat);
  switch (stmat){
  case initial_stiff:
  case tangent_stiff:
  case secant_stiff:
  case incr_tangent_stiff:{
    break;
  }
  case ijth_tangent_stiff:{
    xfscanf (in,"%ld %ld",&ithstiffchange,&jthstiffchange);
    break;
  }
  default:{
    print_err("unknown type of stiffness matrix is required",__FILE__,__LINE__,__func__);
  }
  }

  switch (tnlinsol){
  case arcl:
  case adaptram:
  case arclrv1:
  case arclrv:
  case arclg:{
    if (mespr==1)  fprintf (stdout,"\n system of nonlinear equations will be solved by arc-length method");
    xfscanf (in, "%k%m", "lambda_determination", &detlambda_kwdset, &dlam);
    xfscanf (in, "%k%ld%k%ld", "al_num_steps", &nial, "al_num_iter", &niilal);
    xfscanf (in, "%k%lf%k%lf", "al_error", &erral, "al_init_length", &dlal);
    xfscanf (in, "%k%lf%k%lf", "al_min_length", &dlminal, "al_max_length", &dlmaxal);      
    xfscanf (in, "%k%lf", "al_psi", &psial);
    
    if ((tnlinsol == arclrv1) || (tnlinsol == arclrv))
      xfscanf(in, "%k%lf", "reqval_lambda",&lambdar);
    if (tnlinsol == arclrv)
      xfscanf(in, "%k%lf", "reqerr_lambda",&errl);
    
    if (tnlinsol == arclg){
      // checking of divergency of inner iteration loop
      xfscanf(in, "%k%m", "check_div", &flagsw_kwdset, &check_div);
      if (check_div == on){
	xfscanf(in, "%k%ld", "div_min_steps", &div_min_steps);
	if (div_min_steps < 3){
	  print_warning("the minimum number of performed steps in the divergency check is < 3.\n"
			"It will be set to 3 automatically", __FILE__, __LINE__, __func__);
	  div_min_steps = 3;
	}
      }
      
      // checking of maximum value of the total (cumulative) arc length
      xfscanf(in, "%k%m", "check_total_arcl", &flagsw_kwdset, &check_tot_al);
      if (check_tot_al == on)
	xfscanf(in, "%k%lf", "max_total_arc_length", &max_tot_al);
      // checking of  required value of lambda coefficent for proportional load vector
      xfscanf(in, "%k%m", "check_req_lambda", &flagsw_kwdset, &check_lambdar);
      if (check_lambdar == on){
	xfscanf(in, "%k%lf", "reqval_lambda",&lambdar);
	// tolerance for lambda tresholds
	xfscanf(in, "%k%lf", "reqerr_lambda",&errl);
      }
    }
    
    //fprintf (out,"%ld %d\n",hdbackupal,displnorm);
    //  1 - alldofs
    //  2 - seldofs
    //  3 - seldofscoord
    //  6 - selecnodes
    //  8 - nodesdistincr
    xfscanf (in,"%k%m","al_displ_contr_type",&displacementnorm_kwdset,(int*)&displnorm);
    
    switch (displnorm){
    case alldofs:{  break; }
    case seldofs:{
      xfscanf (in,"%k%ld","num_sel_dofs",&nsdofal);
      selnodal = new long [nsdofal];
      seldofal = new long [nsdofal];
      read_seldof (in);
      break;
    }
    case seldofscoord:{
      xfscanf (in,"%ld",&nsdofal);
      selnodal = new long [nsdofal];
      seldofal = new long [nsdofal];
      selnodcoord = new double [3*nsdofal];
      read_seldofcoord (in);
      break;
    }
    case selecnodes:{
      xfscanf (in,"%k%ld", "num_sel_nodes", &nsnal);
      selnodal = new long [nsnal];
      read_selnod (in);
      break;
    }
    case nodesdistincr:{
      nsnal=2;
      xfscanf (in,"%k%ld","probdimal",&probdimal);
      selnodal = new long [nsnal];
      read_selnod (in);
      break;
    }
    default:
      print_err("unknown norm measurement of displacement increment", __FILE__, __LINE__, __func__);
    }
    
    break;
  }
  case newton:{
    if (mespr==1)  fprintf (stdout,"\n system of nonlinear equations will be solved by Newton-Raphson method");
    xfscanf (in, "%k%ld%k%ld", "nr_num_steps", &ninr, "nr_num_iter", &niilnr);
    xfscanf (in, "%k%lf%k%lf", "nr_error", &errnr, "nr_init_incr", &incrnr);
    xfscanf (in, "%k%lf%k%lf", "nr_minincr", &minincrnr, "nr_maxincr", &maxincrnr);
    break;
  }
  case newtonrv1:{
    if (mespr==1)  fprintf (stdout,"\n system of nonlinear equations will be solved by Newton-Raphson method");
    xfscanf (in, "%k%ld%k%ld", "nr_num_steps", &ninr, "nr_num_iter", &niilnr);
    xfscanf (in, "%k%lf%k%lf", "nr_error", &errnr, "nr_init_incr", &incrnr);
    xfscanf (in, "%k%lf%k%lf", "nr_minincr", &minincrnr, "nr_maxincr", &maxincrnr);
    xfscanf (in, "%k%lf", "req_lambda", &lambdar);
    if (mespr==1)  fprintf(stdout, "\n requried value of lambda=% e", lambdar);
    break;
  }
  /*
  case newtonrestart:{
    if (mespr==1)  fprintf (stdout,"\n system of nonlinear equations will be solved by Newton-Raphson method with ");
    xfscanf (in,"%ld %ld %lf %lf %lf",&ninr,&niilnr,&errnr,&incrnr,&minincrnr);
    xfscanf (in, "%ld %ld", &nienr, &hdbr);
    if (hdbr)
      xfscanf(in, "%s %ld", backupfname, &hdbid);
    if ((mespr==1) && (hdbr))  fprintf(stdout, "\n restart will be performed from file: %s\n record number : %ld\n", backupfname, hdbid);
    break;
  }
  */
  case displctrl:{
    if (mespr==1)  fprintf (stdout,"\n system of nonlinear equations will be solved by Newton-Raphson method with displacement control");
    xfscanf (in, "%k%ld%k%ld", "nr_num_steps", &ninr, "nr_num_iter", &niilnr);
    xfscanf (in, "%k%lf%k%lf", "nr_error", &errnr, "nr_init_incr", &incrnr);
    xfscanf (in, "%k%lf%k%lf", "nr_minincr", &minincrnr, "nr_maxincr", &maxincrnr);
    break;
  }
  case displctrlrv:{
    if (mespr==1)  fprintf (stdout,"\n system of nonlinear equations will be solved by Newton-Raphson method with displacement control");
    xfscanf (in, "%k%ld%k%ld", "nr_num_steps", &ninr, "nr_num_iter", &niilnr);
    xfscanf (in, "%k%lf%k%lf", "nr_error", &errnr, "nr_init_incr", &incrnr);
    xfscanf (in, "%k%lf%k%lf", "nr_minincr", &minincrnr, "nr_maxincr", &maxincrnr);
    xfscanf (in, "%k%lf%k%lf", "req_lambda", &lambdar, "req_lam_err", &errl);
    if (mespr==1)  fprintf(stdout, "\n requried value of lambda=% e, required error of lambda=% e", lambdar, errl);
    break;
  }
  case newtong:
    if (mespr==1)  fprintf (stdout,"\n system of nonlinear equations will be solved by Newton-Raphson method");
    xfscanf (in, "%k%ld%k%ld", "nr_num_steps", &ninr, "nr_num_iter", &niilnr);
    xfscanf (in, "%k%lf%k%lf", "nr_error", &errnr, "nr_init_incr", &incrnr);
    xfscanf (in, "%k%lf%k%lf", "nr_minincr", &minincrnr, "nr_maxincr", &maxincrnr);

    // checking of divergency of inner iteration loop
    xfscanf(in, "%k%m", "check_div", &flagsw_kwdset, &check_div);
    if (check_div == on)
    {
      xfscanf(in, "%k%ld", "div_min_steps", &div_min_steps);
      if (div_min_steps < 3)
      {
        print_warning("the minimum number of performed steps in the divergency check is < 3.\n"
                      "It will be set to 3 automatically", __FILE__, __LINE__, __func__);
        div_min_steps = 3;
      }
    }
  
    // checking of  required value of lambda coefficent for proportional load vector
    xfscanf(in, "%k%m", "check_req_lambda", &flagsw_kwdset, &check_lambdar);
    if (check_lambdar == on)
    {
      xfscanf(in, "%k%lf", "reqval_lambda",&lambdar);
      // tolerance for lambda tresholds
      xfscanf(in, "%k%lf", "reqerr_lambda",&errl);
    }
    break;
  default:
    print_err("unknown solver of nonlinear system of equations is required", __FILE__, __LINE__, __func__);
  }
  if (check_div == on)
  {
    divc_step = new double[div_min_steps];
    divc_err  = new double[div_min_steps];
  }
}



/**
  The function prints parameters for individual types of solvers to the opened text file.
  
  @param out - pointer to the opened text file for output

  @return The function does not return anything.

  Created by Jaroslav Kruis, 2008
  Modified by Tomas Koudelka, 08.2011
*/
void nonlinman::print (FILE *out)
{
  //  type of nonlinear solver
  fprintf (out,"%d\n",(int)tnlinsol);
  
  fprintf (out, "\n%d \n", (int)stmat);
  switch (stmat){
  case initial_stiff:
  case tangent_stiff:
  case secant_stiff:
  case incr_tangent_stiff:{
    break;
  }
  case ijth_tangent_stiff:{
    fprintf (out,"%ld %ld\n",ithstiffchange,jthstiffchange);
    break;
  }
  default:{
    print_err("unknown type of stiffness matrix is required",__FILE__,__LINE__,__func__);
  }
  }

  switch (tnlinsol){
  case arcl:
  case arclrv1:
  case arclg:
  {
    fprintf (out,"%d\n",(int)dlam);
    fprintf (out, "%ld %ld %e %e %e %e %e\n", nial, niilal, erral, dlal,
	     dlminal, dlmaxal, psial);
    if ((tnlinsol == arclrv1) || (tnlinsol == arclrv))
      fprintf(out, "%f\n", lambdar);
    if (tnlinsol == arclrv)
      fprintf(out, "%f\n", errl);
    
    if (tnlinsol == arclg)
    {
      // checking of divergency of inner iteration loop
      fprintf(out, "%d", check_div);
      if (check_div == on)
        fprintf(out, " %ld\n", div_min_steps+1);
      else
        fprintf(out, "\n");
  
      // checking of maximum value of the total (cumulative) arc length
      fprintf(out, "%d", check_tot_al);
      if (check_tot_al == on)
        fprintf(out, " %le\n", max_tot_al);
      else
        fprintf(out, "\n");

      // checking of  required value of lambda coefficent for proportional load vector
      fprintf(out, "%d", check_lambdar);
      if (check_lambdar == on)
        fprintf(out, " %le %le\n", lambdar, errl);
      else
        fprintf(out, "\n");
    }
    //fprintf (out,"%ld %d\n",hdbackupal,displnorm);
    fprintf (out,"%d\n",displnorm);
    
    switch (displnorm){
    case alldofs:{  break; }
    case seldofs:{
      fprintf (out,"%ld\n",nsdofal);
      print_seldof (out);
      break;
    }
    case seldofscoord:{
      fprintf (out,"%ld\n",nsdofal);
      print_seldofcoord (out);
      break;
    }
    case selecnodes:{
      fprintf (out,"%ld\n",nsnal);
      print_selnod (out);
      break;
    }
    case nodesdistincr:{
      nsnal=2;
      fprintf (out,"%ld\n",probdimal);
      print_selnod (out);
      break;
    }
    default:
      print_err("unknown norm measurement of displacement increment", __FILE__, __LINE__, __func__);
    }
    break;
  }
  case newton:{
    fprintf (out, "%ld %ld %e %e %e %e\n", ninr, niilnr, errnr, incrnr, minincrnr, maxincrnr);
    break;      
  }
  case newtonrv1:{
    fprintf (out, "%ld %ld %e %e %e %e %e\n", ninr, niilnr, errnr, incrnr, minincrnr, maxincrnr, lambdar);
    break;
  }
  case newtonrestart:{
    fprintf (out,"%ld %ld %e %e %e\n",ninr,niilnr,errnr,incrnr,minincrnr);
    fprintf (out, "%ld %ld\n", nienr, hdbr);
    if (hdbr)
      fprintf(out, "%s\n %ld\n", backupfname, hdbid);
    break;
  }
  case displctrl:{
    fprintf (out, "%ld %ld %e %e %e\n", ninr, niilnr, errnr, incrnr, minincrnr);
    break;      
  }
  case displctrlrv:{
    fprintf (out,"%ld %ld %lf %lf %lf %lf %lf\n",ninr,niilnr,errnr,incrnr,minincrnr,lambdar,errl);
    break;
  }
  case newtong:
  {
    fprintf (out, "%ld %ld\n", ninr, niilnr);
    fprintf (out, "%le %le %le %le\n", errnr, incrnr, minincrnr, maxincrnr);
    // checking of divergency of inner iteration loop
    fprintf(out, "%d", check_div);
    if (check_div == on)
      fprintf(out, " %ld\n", div_min_steps+1);
    else
      fprintf(out, "\n");
    // checking of  required value of lambda coefficent for proportional load vector
    fprintf(out, "%d", check_lambdar);
    if (check_lambdar == on)
      fprintf(out, " %le %le\n", lambdar, errl);
    else
      fprintf(out, "\n");

    break;
  }
  default:{}
  }
  
}



/**
  The function reads selected nodes for arclength control.

  @param in - pointer to the opened text file

  @return The function does not return anything.

  Created by Jaroslav Kruis, 2008
*/
void nonlinman::read_selnod (XFILE *in)
{
  long i;
  
  for (i=0;i<nsnal;i++){
    xfscanf (in,"%ld",selnodal+i);
    selnodal[i]--;
  }
  
}



/**
  The function reads selected dofs for arclength control.

  @param in - pointer to the opened text file

  @return The function does not return anything.

  Created by Jaroslav Kruis, 2008
*/
void nonlinman::read_seldof (XFILE *in)
{
  long i;
  
  for (i=0;i<nsdofal;i++){
    xfscanf (in,"%ld %ld",selnodal+i,seldofal+i);
    selnodal[i]--;  seldofal[i]--;
  }
  
}



/**
  The function reads coordinates of selected dofs for arclength control.

  @param in - pointer to the opened text file

  @return The function does not return anything.

  Created by Jaroslav Kruis, 2008
*/
void nonlinman::read_seldofcoord (XFILE *in)
{
  long i;

  for (i=0;i<nsdofal;i++){
    xfscanf (in,"%lf %lf %lf %ld",selnodcoord+3*i,selnodcoord+3*i+1,selnodcoord+3*i+2,seldofal+i);
    seldofal[i]--;
  }
}



/**
  The function prints selected nodes for arclength control.

  @param in - pointer to the opened text file

  @return The function does not return anything.

  Created by Jaroslav Kruis, 2008
*/
void nonlinman::print_selnod (FILE *out)
{
  long i;
  
  for (i=0;i<nsnal;i++){
    fprintf (out,"%ld\n",selnodal[i]+1);
  }
  
}



/**
  The function prints selected dofs for arclength control.

  @param in - pointer to the opened text file

  @return The function does not return anything.

  Created by Jaroslav Kruis, 2008
*/
void nonlinman::print_seldof (FILE *out)
{
  long i;
  
  for (i=0;i<nsdofal;i++){
    fprintf (out,"%ld %ld\n",selnodal[i]+1,seldofal[i]+1);
  }
  
}



/**
  The function prints coordinates of selected dofs for arclength control.

  @param in - pointer to the opened text file

  @return The function does not return anything.

  Created by Jaroslav Kruis, 2008
*/
void nonlinman::print_seldofcoord (FILE *out)
{
  long i;
  
  for (i=0;i<nsdofal;i++){
    fprintf (out,"%f %f %f %ld\n",selnodcoord[3*i],selnodcoord[3*i+1],selnodcoord[3*i+2],seldofal[i]+1);
  }
  
}



/**
  The function initializes data from the another nonlinman structure.

  @param nm - nonlinman structure used fro initialization

  @raturn The function does not return anything.

  Created by Jaroslav Kruis, 2008
*/
void nonlinman::initiate (nonlinman &nm)
{
  long i;

  tnlinsol = nm.tnlinsol;

  stmat = nm.stmat;
  ithstiffchange = nm.ithstiffchange;
  jthstiffchange = nm.jthstiffchange;

  //  ARC-LENGTH METHOD
  displnorm = nm.displnorm;
  dlam = nm.dlam;
  hdbackupal = nm.hdbackupal;
  nial = nm.nial;
  niilal = nm.niilal;
  erral = nm.erral;
  dlal = nm.dlal;
  dlminal = nm.dlminal;
  dlmaxal = nm.dlmaxal;
  psial = nm.psial;
  
  nsnal = nm.nsnal;
  nsdofal = nm.nsdofal;
  probdimal = nm.probdimal;
  check_tot_al = nm.check_tot_al;
  max_tot_al = nm.max_tot_al;

  nxal = nm.nxal;
  nyal = nm.nyal;
  nzal = nm.nzal;
  
  
  switch (displnorm){
  case alldofs:{  break; }
  case seldofs:{
    selnodal = new long [nsdofal];
    seldofal = new long [nsdofal];
    
    for (i=0;i<nsdofal;i++){
      selnodal[i]=nm.selnodal[i];
      seldofal[i]=nm.seldofal[i];
    }
    
    break;
  }
  case seldofscoord:{
    seldofal = new long [nsdofal];
    selnodcoord = new double [3*nsdofal];
    
    for (i=0;i<nsdofal;i++){
      selnodcoord[3*i]=nm.selnodcoord[3*i];
      selnodcoord[3*i+1]=nm.selnodcoord[3*i+1];
      selnodcoord[3*i+2]=nm.selnodcoord[3*i+2];
      seldofal[i]=nm.seldofal[i];
    }
    
    break;
  }
  case selecnodes:{
    selnodal = new long [nsnal];
    
    for (i=0;i<nsnal;i++){
      selnodal[i]=nm.selnodal[i];
    }
    
    break;
  }
  case nodesdistincr:{
    nsnal=2;
    selnodal = new long [nsnal];
    
    for (i=0;i<nsnal;i++){
      selnodal[i]=nm.selnodal[i];
    }
    
    break;
  }
  default:
    print_err("unknown norm measurement of displacement increment", __FILE__, __LINE__, __func__);
  }
  
  // COMMON SETUP for both Newton-Raphson and arclength methods
  check_lambdar = nm.check_lambdar;
  lambdar = nm.check_lambdar;
  errl = nm.errl;

  check_div = nm.check_div;
  div_min_steps = nm.div_min_steps;
  if (check_div == on)
  {
    divc_step = new double[div_min_steps];
    divc_err = new double[div_min_steps]; 
    for(i=0; i<div_min_steps; i++)
    {
      divc_step[i] = nm.divc_step[i];
      divc_err[i] = nm.divc_err[i];
    }
  }
  
  //  NEWTON-RAPHSON METHOD
  ninr = nm.ninr;
  niilnr = nm.niilnr;
  errnr = nm.errnr;
  incrnr = nm.incrnr;
  minincrnr = nm.minincrnr;
  maxincrnr = nm.maxincrnr;
  lambdar = nm.lambdar;
  nienr = nm.nienr;
  hdbr  = nm.hdbr;
  if (hdbr)
  {
    hdbid = nm.hdbid;
    for (i=0; i < long(strlen(nm.backupfname)); i++)
      backupfname[i] = nm.backupfname[i];
  }
}
